from typing import Any, Callable, Dict, Tuple

from dash import Dash, Input, Output

from webviz_subsurface._models.well_set_model import WellSetModel

from ..utils.xtgeo_well_log_to_json import xtgeo_well_logs_to_json_format


def well_controller(
    app: Dash,
    well_set_model: WellSetModel,
    log_templates: Dict,
    get_uuid: Callable,
) -> None:
    @app.callback(
        Output(get_uuid("well-log-viewer"), "welllog"),
        Output(get_uuid("well-log-viewer"), "template"),
        Input(get_uuid("well"), "value"),
        Input(get_uuid("template"), "value"),
    )
    def _update_log_data(well_name: str, template: str) -> Tuple[Any, Any]:
        well = well_set_model.get_well(well_name)
        return xtgeo_well_logs_to_json_format(well), log_templates.get(template)



from typing import Optional


def add_vector_to_vector_selector_data(
    vector_selector_data: list,
    vector: str,
    description: Optional[str] = None,
    description_at_last_node: bool = False,
) -> None:
    nodes = vector.split(":")
    current_child_list = vector_selector_data
    for index, node in enumerate(nodes):
        found = False
        for child in current_child_list:
            if child["name"] == node:
                children = child["children"]
                current_child_list = children if children is not None else []
                found = True
                break
        if not found:
            node_data: dict = {
                "name": node,
                "children": [] if index < len(nodes) - 1 else None,
            }
            if not description_at_last_node and description and index == 0:
                node_data["description"] = description
            elif description_at_last_node and description and (index == len(nodes) - 1):
                node_data["description"] = description

            current_child_list.append(node_data)

            children = current_child_list[-1]["children"]
            current_child_list = children if children is not None else []


def is_vector_name_in_vector_selector_data(
    name: str, vector_selector_data: list
) -> bool:
    nodes = name.split(":")
    current_child_list = vector_selector_data
    for node in nodes:
        found = False
        for child in current_child_list:
            if child["name"] == node:
                children = child["children"]
                current_child_list = children if children is not None else []
                found = True
                break
        if not found:
            return False
    return found



"""These are plugins relevant within subsurface workflows. Most of them
rely on the setting `scratch_ensembles` within
`shared_settings`. This setting connects ensemble names (user defined)
with the paths to where the ensembles are stored, either absolute or
relative to the location of the configuration file.
I.e. you could have
```yaml
title: Reek Webviz Demonstration
shared_settings:
  scratch_ensembles:
    iter-0: /scratch/my_ensemble/realization-*/iter-0
    iter-1: /scratch/my_ensemble/realization-*/iter-1
pages:
  - title: Front page
    content:
      - plugin: ReservoirSimulationTimeSeries
        ensembles:
          - iter-0
          - iter-1
```
"""



import json
from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple, Union
from uuid import uuid4

import numpy as np
import webviz_core_components as wcc
from dash import Dash, Input, Output, State, callback_context, dcc, html
from dash.exceptions import PreventUpdate
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import calculate_slider_step
from webviz_config.webviz_store import webvizstore

from .._datainput.seismic import get_iline, get_xline, get_zslice, load_cube_data


class SegyViewer(WebvizPluginABC):
    """Inspired by [SegyViewer for Python](https://github.com/equinor/segyviewer) this plugin
visualizes seismic 3D cubes with 3 plots (inline, crossline and zslice).
The plots are linked and updates are done by clicking in the plots.

---

* **`segyfiles`:** List of file paths to `SEGY` files (absolute or relative to config file).
* **`zunit`:** z-unit for display.
* **`colors`:** List of hex colors use. \
Note that apostrophies should be used to avoid that hex colors are read as comments. E.g. \
`'#000000'` for black.

---

* [Examples of segyfiles](https://github.com/equinor/webviz-subsurface-testdata/tree/master/\
observed_data/seismic).

The segyfiles are on a `SEG-Y` format and can be investigated outside `webviz` using \
e.g. [xtgeo](https://xtgeo.readthedocs.io/en/latest/).

"""

    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        segyfiles: List[Path],
        zunit: str = "depth (m)",
        colors: list = None,
    ):

        super().__init__()

        self.zunit = zunit
        self.segyfiles: List[str] = [str(segy) for segy in segyfiles]
        self.initial_colors = (
            colors
            if colors
            else [
                "#67001f",
                "#ab152a",
                "#d05546",
                "#ec9372",
                "#fac8ac",
                "#faeae1",
                "#e6eff4",
                "#bbd9e9",
                "#7db7d6",
                "#3a8bbf",
                "#1f61a5",
                "#053061",
            ]
        )
        self.init_state = self.update_state(self.segyfiles[0])
        self.init_state.get("colorscale", self.initial_colors)
        self.init_state.get("uirevision", str(uuid4()))

        self.plotly_theme = webviz_settings.theme.plotly_theme
        self.set_callbacks(app)

    def update_state(self, cubepath: str, **kwargs: Any) -> Dict[str, Any]:
        cube = load_cube_data(get_path(cubepath))
        state = {
            "cubepath": cubepath,
            "iline": int(cube.ilines[int(len(cube.ilines) / 2)]),
            "xline": int(cube.xlines[int(len(cube.xlines) / 2)]),
            "zslice": float(cube.zslices[int(len(cube.zslices) / 2)]),
            "min_value": float(f"{round(cube.values.min(), 2):2f}"),
            "max_value": float(f"{round(cube.values.max(), 2):2f}"),
            "color_min_value": float(f"{round(cube.values.min(), 2):2f}"),
            "color_max_value": float(f"{round(cube.values.max(), 2):2f}"),
            "uirevision": str(uuid4()),
            "colorscale": self.initial_colors,
        }
        if kwargs:
            for key, value in kwargs.items():
                state[key] = value
        return state

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("layout"),
                "content": (
                    "Visualizes SEG-Y cubes. Display different slices "
                    "of the cube by clicking (MB1) in the different plots. "
                    "To zoom, hold MB1 and draw a vertical/horizontal "
                    "line or a rectangle."
                ),
            },
            {
                "id": self.uuid("cube"),
                "content": "The currently visualized seismic cube.",
            },
            {
                "id": self.uuid("inline"),
                "content": "Selected inline for the seismic cube.",
            },
            {
                "id": self.uuid("xline"),
                "content": "Selected crossline for the seismic cube.",
            },
            {
                "id": self.uuid("zslice"),
                "content": "Selected zslice for the seismic cube.",
            },
            {
                "id": self.uuid("color-scale"),
                "content": "Click this button to change colorscale",
            },
            {
                "id": self.uuid("color-values"),
                "content": "Drag either node of slider to truncate color ranges.",
            },
            {
                "id": self.uuid("color-reset"),
                "content": (
                    "Click this button to update color slider min/max and reset ranges."
                ),
            },
            {
                "id": self.uuid("zoom"),
                "content": "Click this button to reset zoom/pan state in the plot.",
            },
        ]

    @property
    def settings_layout(self) -> wcc.FlexBox:
        """Layout for color and other settings"""
        return wcc.Frame(
            style={"width": "40%"},
            children=[
                html.Div(
                    style={"width": "100%"},
                    children=wcc.Dropdown(
                        label="Seismic cube",
                        id=self.uuid("cube"),
                        options=[
                            {"label": Path(cube).stem, "value": cube}
                            for cube in self.segyfiles
                        ],
                        value=self.segyfiles[0],
                        clearable=False,
                    ),
                ),
                html.Div(
                    children=[
                        wcc.Label(
                            children="Set colorscale",
                        ),
                        wcc.ColorScales(
                            id=self.uuid("color-scale"),
                            colorscale=self.initial_colors,
                            nSwatches=12,
                        ),
                    ],
                ),
                html.Div(
                    children=[
                        wcc.RangeSlider(
                            label="Color range",
                            id=self.uuid("color-values"),
                            min=self.init_state["min_value"],
                            max=self.init_state["max_value"],
                            value=[
                                self.init_state["min_value"],
                                self.init_state["max_value"],
                            ],
                            tooltip={"placement": "bottom"},
                            step=calculate_slider_step(
                                min_value=self.init_state["min_value"],
                                max_value=self.init_state["max_value"],
                                steps=100,
                            ),
                        ),
                    ],
                ),
                html.Button(id=self.uuid("color-reset"), children="Reset Range"),
                html.Button(id=self.uuid("zoom"), children="Reset zoom"),
            ],
        )

    @property
    def layout(self) -> html.Div:
        return html.Div(
            id=self.uuid("layout"),
            children=wcc.FlexBox(
                children=[
                    wcc.FlexColumn(
                        children=[
                            wcc.Frame(
                                color="white",
                                style={"minWidth": "200px", "height": "400px"},
                                children=wcc.Graph(
                                    config={"displayModeBar": False},
                                    id=self.uuid("inline"),
                                ),
                            ),
                            wcc.Frame(
                                color="white",
                                style={"minWidth": "200px", "height": "400px"},
                                children=wcc.Graph(
                                    config={"displayModeBar": False},
                                    id=self.uuid("xline"),
                                ),
                            ),
                        ],
                    ),
                    wcc.FlexColumn(
                        children=[
                            wcc.Frame(
                                color="white",
                                style={"minWidth": "200px", "height": "400px"},
                                children=wcc.Graph(
                                    config={"displayModeBar": False},
                                    id=self.uuid("zslice"),
                                ),
                            ),
                            html.Div(
                                style={"height": "400px"}, children=self.settings_layout
                            ),
                        ],
                    ),
                    dcc.Store(
                        id=self.uuid("state-storage"),
                        storage_type="session",
                        data=json.dumps(self.init_state),
                    ),
                ]
            ),
        )

    # pylint: disable=too-many-statements
    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self.uuid("state-storage"), "data"),
            [
                Input(self.uuid("cube"), "value"),
                Input(self.uuid("inline"), "clickData"),
                Input(self.uuid("xline"), "clickData"),
                Input(self.uuid("zslice"), "clickData"),
                Input(self.uuid("color-values"), "value"),
                Input(self.uuid("color-scale"), "colorscale"),
                Input(self.uuid("zoom"), "n_clicks"),
                Input(self.uuid("color-reset"), "n_clicks"),
            ],
            [
                State(self.uuid("zslice"), "figure"),
                State(self.uuid("state-storage"), "data"),
            ],
        )
        def _update_state(
            cubepath: str,
            icd: Union[dict, None],
            xcd: Union[dict, None],
            zcd: Union[dict, None],
            color_values: List[float],
            colorscale: List[float],
            _zoom_btn: Union[int, None],
            _reset_range_btn: Union[int, None],
            zfig: Union[dict, None],
            state_data_str: str,
        ) -> str:
            """Updates dcc.Store object with active iline, xline, zlayer and color settings."""

            store: dict = json.loads(state_data_str)
            ctx = callback_context.triggered[0]["prop_id"]
            x_was_clicked = (
                xcd if xcd and ctx == f"{self.uuid('xline')}.clickData" else None
            )
            i_was_clicked = (
                icd if icd and ctx == f"{self.uuid('inline')}.clickData" else None
            )
            z_was_clicked = (
                zcd if zcd and ctx == f"{self.uuid('zslice')}.clickData" else None
            )
            if ctx == f"{self.uuid('cube')}.value":
                store = self.update_state(cubepath)
            if ctx == f"{self.uuid('zoom')}.n_clicks":
                store["uirevision"] = str(uuid4())
            if x_was_clicked and xcd is not None:
                store["iline"] = xcd["points"][0]["x"]
                store["zslice"] = xcd["points"][0]["y"]
            if z_was_clicked and zcd is not None and zfig is not None:
                store["iline"] = zcd["points"][0]["x"]
                store["xline"] = zcd["points"][0]["y"]
                store["zslice"] = float(zfig["data"][0]["text"])
            if i_was_clicked and icd is not None:
                store["xline"] = icd["points"][0]["x"]
                store["zslice"] = icd["points"][0]["y"]
            if ctx == f"{self.uuid('color-reset')}.n_clicks":
                store["color_min_value"] = store["min_value"]
                store["color_max_value"] = store["max_value"]
            else:
                store["color_min_value"] = color_values[0]
                store["color_max_value"] = color_values[1]
            store["colorscale"] = colorscale
            return json.dumps(store)

        @app.callback(
            Output(self.uuid("zslice"), "figure"),
            [Input(self.uuid("state-storage"), "data")],
        )
        def _set_zslice(state_data_str: Union[str, None]) -> dict:
            """Updates z-slice heatmap"""
            if not state_data_str:
                raise PreventUpdate
            state = json.loads(state_data_str)
            cube = load_cube_data(get_path(state["cubepath"]))
            shapes = [
                {
                    "type": "line",
                    "x0": state["iline"],
                    "y0": cube.xlines[0],
                    "x1": state["iline"],
                    "y1": cube.xlines[-1],
                    "line": {"width": 1, "dash": "dot"},
                },
                {
                    "type": "line",
                    "y0": state["xline"],
                    "x0": cube.ilines[0],
                    "y1": state["xline"],
                    "x1": cube.ilines[-1],
                    "line": {"width": 1, "dash": "dot"},
                },
            ]

            zslice_arr = get_zslice(cube, state["zslice"])

            fig = make_heatmap(
                zslice_arr,
                self.plotly_theme,
                xaxis=cube.ilines,
                yaxis=cube.xlines,
                showscale=True,
                text=str(state["zslice"]),
                title=f'Zslice {state["zslice"]} ({self.zunit})',
                xaxis_title="Inline",
                yaxis_title="Crossline",
                reverse_y=False,
                zmin=state["color_min_value"],
                zmax=state["color_max_value"],
                colorscale=state["colorscale"],
                uirevision=state["uirevision"],
            )
            fig["layout"]["shapes"] = shapes
            fig["layout"]["xaxis"].update({"constrain": "domain"})
            fig["layout"]["yaxis"].update({"scaleanchor": "x"})
            return fig

        @app.callback(
            Output(self.uuid("inline"), "figure"),
            [Input(self.uuid("state-storage"), "data")],
        )
        def _set_iline(state_data_str: Union[str, None]) -> dict:
            """Updates inline heatmap"""
            if not state_data_str:
                raise PreventUpdate
            state = json.loads(state_data_str)
            cube = load_cube_data(get_path(state["cubepath"]))
            shapes = [
                {
                    "type": "line",
                    "x0": state["xline"],
                    "y0": cube.zslices[0],
                    "x1": state["xline"],
                    "y1": cube.zslices[-1],
                    "line": {"width": 1, "dash": "dot"},
                },
                {
                    "type": "line",
                    "x0": cube.xlines[0],
                    "y0": state["zslice"],
                    "x1": cube.xlines[-1],
                    "y1": state["zslice"],
                    "line": {"width": 1, "dash": "dot"},
                },
            ]
            iline_arr = get_iline(cube, state["iline"])

            fig = make_heatmap(
                iline_arr,
                self.plotly_theme,
                xaxis=cube.xlines,
                yaxis=cube.zslices,
                title=f'Inline {state["iline"]}',
                xaxis_title="Crossline",
                yaxis_title=self.zunit,
                zmin=state["color_min_value"],
                zmax=state["color_max_value"],
                colorscale=state["colorscale"],
                uirevision=state["uirevision"],
            )
            fig["layout"]["shapes"] = shapes
            return fig

        @app.callback(
            Output(self.uuid("xline"), "figure"),
            [Input(self.uuid("state-storage"), "data")],
        )
        def _set_xline(state_data_str: Union[str, None]) -> dict:
            """Update xline heatmap"""
            if not state_data_str:
                raise PreventUpdate
            state = json.loads(state_data_str)
            cube = load_cube_data(get_path(state["cubepath"]))
            shapes = [
                {
                    "type": "line",
                    "x0": state["iline"],
                    "y0": cube.zslices[0],
                    "x1": state["iline"],
                    "y1": cube.zslices[-1],
                    "line": {"width": 1, "dash": "dot"},
                },
                {
                    "type": "line",
                    "x0": cube.ilines[0],
                    "y0": state["zslice"],
                    "x1": cube.ilines[-1],
                    "y1": state["zslice"],
                    "line": {"width": 1, "dash": "dot"},
                },
            ]
            xline_arr = get_xline(cube, state["xline"])
            fig = make_heatmap(
                xline_arr,
                self.plotly_theme,
                xaxis=cube.ilines,
                yaxis=cube.zslices,
                title=f'Crossline {state["xline"]}',
                xaxis_title="Inline",
                yaxis_title=self.zunit,
                zmin=state["color_min_value"],
                zmax=state["color_max_value"],
                colorscale=state["colorscale"],
                uirevision=state["uirevision"],
            )
            fig["layout"]["shapes"] = shapes
            return fig

        @app.callback(
            [
                Output(self.uuid("color-values"), "min"),
                Output(self.uuid("color-values"), "max"),
                Output(self.uuid("color-values"), "value"),
                Output(self.uuid("color-values"), "step"),
            ],
            [Input(self.uuid("color-reset"), "n_clicks")],
            [State(self.uuid("state-storage"), "data")],
        )
        def _reset_color_range(
            _btn_clicked: int, state_data_str: Union[str, None]
        ) -> Tuple[float, float, list, float]:
            if not state_data_str:
                raise PreventUpdate
            state = json.loads(state_data_str)
            minv = state["min_value"]
            maxv = state["max_value"]
            value = [minv, maxv]
            step = calculate_slider_step(min_value=minv, max_value=maxv, steps=100)
            return minv, maxv, value, step

    @staticmethod
    def set_grid_layout(columns: str) -> dict:
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
        }

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [(get_path, [{"path": fn}]) for fn in self.segyfiles]


# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def make_heatmap(
    arr: np.ndarray,
    theme: dict,
    height: int = 400,
    zmin: float = None,
    zmax: float = None,
    colorscale: list = None,
    uirevision: str = None,
    showscale: bool = False,
    reverse_y: bool = True,
    xaxis: np.ndarray = None,
    yaxis: np.ndarray = None,
    text: str = None,
    title: str = None,
    yaxis_title: str = None,
    xaxis_title: str = None,
) -> Dict[str, Any]:
    """Createst heatmap plot"""
    colors = (
        [[i / (len(colorscale) - 1), color] for i, color in enumerate(colorscale)]
        if colorscale
        else "RdBu"
    )
    layout = {}
    layout.update(theme["layout"])
    layout.update(
        {
            "height": height,
            "title": title,
            "uirevision": uirevision,
            "margin": {"b": 50, "t": 50, "r": 0},
            "yaxis": {
                "title": yaxis_title,
                "autorange": "reversed" if reverse_y else None,
            },
            "xaxis": {"title": xaxis_title},
        }
    )

    return {
        "data": [
            {
                "type": "heatmap",
                "text": text if text else None,
                "z": arr.tolist(),
                "x": xaxis,
                "y": yaxis,
                "zsmooth": "best",
                "showscale": showscale,
                "colorscale": colors,
                "zmin": zmin,
                "zmax": zmax,
            }
        ],
        "layout": layout,
    }


@webvizstore
def get_path(path: str) -> Path:
    return Path(path)



from dataclasses import dataclass

from .ensemble_surface_provider import SurfaceAddress


@dataclass(frozen=True)
class QualifiedSurfaceAddress:
    provider_id: str
    address: SurfaceAddress


@dataclass(frozen=True)
class QualifiedDiffSurfaceAddress:
    provider_id_a: str
    address_a: SurfaceAddress
    provider_id_b: str
    address_b: SurfaceAddress



from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from ...._types import CorrType, DepthType


class Options(SettingsGroupABC):
    class Ids(StrEnum):
        CORRTYPE = "corrtype"
        DEPTHTYPE = "depthtype"

    def __init__(self) -> None:
        super().__init__("Options")

    def layout(self) -> List[Component]:
        return [
            wcc.RadioItems(
                label="Correlation options",
                id=self.register_component_unique_id(self.Ids.CORRTYPE),
                options=[
                    {
                        "label": "Simulated vs parameters",
                        "value": CorrType.SIM_VS_PARAM,
                    },
                    {
                        "label": "Parameter vs simulated",
                        "value": CorrType.PARAM_VS_SIM,
                    },
                ],
                value="sim_vs_param",
            ),
            wcc.RadioItems(
                label="Depth option",
                id=self.register_component_unique_id(self.Ids.DEPTHTYPE),
                options=[
                    {
                        "label": "TVD",
                        "value": DepthType.TVD,
                    },
                    {
                        "label": "MD",
                        "value": DepthType.MD,
                    },
                ],
                value=DepthType.TVD,
            ),
        ]



from dash import html
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewElementABC


class WellCompletionViewElement(ViewElementABC):
    class Ids(StrEnum):
        COMPONENT = "component"

    def __init__(self) -> None:
        super().__init__()

    def inner_layout(self) -> html.Div:
        return html.Div(
            id=self.register_component_unique_id(
                WellCompletionViewElement.Ids.COMPONENT
            )
        )



import numpy as np
import pandas as pd


class ParametersModel:
    """Class to process ensemble parameter data"""

    POSSIBLE_SELECTORS = [
        "ENSEMBLE",
        "REAL",
        "SENSNAME",
        "SENSCASE",
        "SENSTYPE",
        "SENSNAME_CASE",
    ]

    def __init__(
        self,
        dataframe: pd.DataFrame,
        drop_constants: bool = True,
        keep_numeric_only: bool = True,
        drop_parameters_with_nan: bool = False,
    ) -> None:
        self._dataframe = dataframe if dataframe is not None else pd.DataFrame()
        self._validate_dframe()
        self._sensrun = self._check_if_sensitivity_run()
        self._prepare_data(drop_constants, keep_numeric_only, drop_parameters_with_nan)
        self._parameters = [
            x for x in self._dataframe if x not in self.POSSIBLE_SELECTORS
        ]
        self._parameters_per_ensemble = self._split_parameters_by_ensemble()

    @property
    def ensembles(self) -> list:
        return list(self._dataframe["ENSEMBLE"].unique())

    @property
    def parameters(self) -> list:
        return self._parameters

    @property
    def parameters_per_ensemble(self) -> dict:
        return self._parameters_per_ensemble

    @property
    def selectors(self) -> list:
        return [col for col in self.POSSIBLE_SELECTORS if col in self.dataframe]

    @property
    def sensitivities(self) -> list:
        return list(self._dataframe["SENSNAME"].unique()) if self.sensrun else []

    @property
    def sensrun(self) -> bool:
        return self._sensrun

    @property
    def sensitivity_ensembles(self) -> list:
        return self._sensitivity_ensembles

    @property
    def mc_ensembles(self) -> list:
        return [ens for ens in self.ensembles if ens not in self._sensitivity_ensembles]

    @property
    def sens_df(self) -> pd.DataFrame:
        return self.dataframe[self.POSSIBLE_SELECTORS]

    @property
    def dataframe(self) -> pd.DataFrame:
        return self._dataframe

    def _validate_dframe(self) -> None:
        for col in ["REAL", "ENSEMBLE"]:
            if col not in self._dataframe:
                raise KeyError(f"Required columns {col} not found in dataframe")

    def _prepare_data(
        self,
        drop_constants: bool,
        keep_numeric_only: bool,
        drop_parameters_with_nan: bool,
    ) -> None:
        """
        Different data preparations on the parameters, before storing them as an attribute.
        Option to drop parameters with constant values. Prefixes on parameters from GEN_KW
        are removed, in addition parameters with LOG distribution will be kept while the
        other is dropped.
        """

        # Remove parameters with only NaN (can happen for filtered dataframes)
        self._dataframe = self._dataframe.dropna(axis=1, how="all")

        if drop_constants:
            constant_params = [
                param
                for param in self._dataframe
                if self._dataframe[param].dropna().nunique() == 1
                and param not in self.POSSIBLE_SELECTORS
            ]
            self._dataframe = self._dataframe.drop(columns=constant_params)

        # Keep only LOG parameters
        log_params = [param for param in self._dataframe if param.startswith("LOG10_")]
        self._dataframe = self._dataframe.drop(
            columns=[param.replace("LOG10_", "") for param in log_params]
        )
        self._dataframe = self._dataframe.rename(
            columns={col: f"{col} (log)" for col in log_params}
        )
        # Remove prefix on parameter name added by GEN_KW
        self._dataframe = self._dataframe.rename(
            columns={
                col: (col.split(":", 1)[1]) for col in self._dataframe if ":" in col
            }
        )
        # Drop columns if duplicate names
        self._dataframe = self._dataframe.loc[:, ~self._dataframe.columns.duplicated()]

        # Only use numeric columns
        if keep_numeric_only:
            numeric_columns = self._dataframe.select_dtypes(
                include=np.number
            ).columns.tolist()
            self._dataframe = self._dataframe[
                self.selectors
                + [col for col in numeric_columns if col not in self.selectors]
            ]

        if drop_parameters_with_nan:
            self._dataframe = self._dataframe.drop(columns=self._get_columns_with_nan())

    def _check_if_sensitivity_run(self) -> bool:
        """
        Flag ensembles as sensrun if more than one sensitivity exists or
        there exist only one sensitivity which is not of type montecarlo.
        """
        self._sensitivity_ensembles = []

        if "SENSNAME" not in self._dataframe:
            return False

        # if mix of gen_kw and sensitivity ensembles add
        # dummy sensitivvity columns to gen_kw ensembles
        gen_kw_mask = self._dataframe["SENSNAME"].isnull()
        self._dataframe.loc[gen_kw_mask, "SENSNAME"] = "🎲"
        self._dataframe.loc[gen_kw_mask, "SENSCASE"] = "p10_p90"

        # set senstype from senscase
        mc_mask = self._dataframe["SENSCASE"] == "p10_p90"
        self._dataframe["SENSTYPE"] = np.where(mc_mask, "mc", "scalar")

        # make combination column of sensname and senscase
        self._dataframe["SENSNAME_CASE"] = np.where(
            mc_mask,
            self._dataframe["SENSNAME"],
            self._dataframe[["SENSNAME", "SENSCASE"]].agg("--".join, axis=1),
        )

        self._sensitivity_ensembles = [
            ens
            for ens, df in self.dataframe.groupby("ENSEMBLE")
            if not (df["SENSNAME"].nunique() == 1 and df["SENSTYPE"].unique() == ["mc"])
        ]
        return bool(self._sensitivity_ensembles)

    def _get_columns_with_nan(self) -> list:
        """Return parameters that contains nan values within ensembles"""
        cols_with_nan = []
        for _, ensdf in self._dataframe.groupby("ENSEMBLE"):
            cols_with_nan.extend(
                [
                    col
                    for col in ensdf
                    if ensdf[col].isnull().values.any()
                    and not ensdf[col].isnull().values.all()
                ]
            )
        return cols_with_nan

    def _split_parameters_by_ensemble(self) -> dict:
        return {
            ens: [
                col
                for col in ensdf.dropna(axis=1, how="all").columns
                if col in self._parameters
            ]
            for ens, ensdf in self._dataframe.groupby("ENSEMBLE")
        }

    def get_parameters_for_ensembles(self, ensembles: list) -> list:
        """Get the unique parameters belonging to one or more ensembles"""
        parameters = set()
        for ens in ensembles:
            parameters.update(self._parameters_per_ensemble[ens])
        return sorted(list(parameters))



from pathlib import Path
from typing import Callable, List, Tuple
from uuid import uuid4

import pandas as pd
from dash import Dash, Input, Output, dcc, html
from webviz_config import WebvizPluginABC
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore
from webviz_subsurface_components import Morris


class MorrisPlot(WebvizPluginABC):
    """Renders a visualization of the Morris sampling method.
The Morris method can be used to screen parameters for how they
influence model response, both individually and through interaction
effect with other parameters.

---

* **`csv_file`:** Input data on csv format.

---

[Example of input file](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
aggregated_data/morris.csv).
"""

    def __init__(self, app: Dash, csv_file: Path):

        super().__init__()

        self.graph_id = f"graph-{uuid4()}"
        self.vector_id = f"vector-{uuid4()}"
        self.csv_file = csv_file
        self.data = read_csv(self.csv_file)
        self.vector_names = self.data["name"].unique()
        self.set_callbacks(app)

    @property
    def layout(self) -> html.Div:
        return html.Div(
            [
                html.Label("Vector", style={"font-size": "2rem"}),
                dcc.Dropdown(
                    id=self.vector_id,
                    clearable=False,
                    options=[{"label": i, "value": i} for i in list(self.vector_names)],
                    value=self.vector_names[0],
                ),
                Morris(id=self.graph_id),
            ]
        )

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [(read_csv, [{"csv_file": self.csv_file}])]

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            [
                Output(self.graph_id, "output"),
                Output(self.graph_id, "parameter"),
                Output(self.graph_id, "parameters"),
            ],
            [Input(self.vector_id, "value")],
        )
        def _update_plot(vector: str) -> Tuple[list, str, List[dict]]:
            df = self.data[self.data["name"] == vector]
            df = df.sort_values("time")
            output = (
                df[["mean", "max", "min", "time"]]
                .drop_duplicates()
                .to_dict(orient="records")
            )
            parameters = []

            for name in self.data["name"].unique():
                if name != vector:
                    name_df = self.data[self.data["name"] == name]
                    parameters.append(
                        {
                            "main": list(name_df["morris_main"]),
                            "name": str(name),
                            "interactions": list(name_df["morris_interaction"]),
                        }
                    )
            return output, vector, parameters


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file)



import logging
import shutil
from enum import Enum
from pathlib import Path
from typing import List, Optional

import pandas as pd
import xtgeo

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._fault_polygons_discovery import FaultPolygonsFileInfo
from .ensemble_fault_polygons_provider import (
    EnsembleFaultPolygonsProvider,
    FaultPolygonsAddress,
    SimulatedFaultPolygonsAddress,
)

LOGGER = logging.getLogger(__name__)

REL_SIM_DIR = "sim"


# pylint: disable=too-few-public-methods
class Col:
    TYPE = "type"
    REAL = "real"
    ATTRIBUTE = "attribute"
    NAME = "name"
    ORIGINAL_PATH = "original_path"
    REL_PATH = "rel_path"


class FaultPolygonsType(str, Enum):
    SIMULATED = "simulated"


class ProviderImplFile(EnsembleFaultPolygonsProvider):
    def __init__(
        self,
        provider_id: str,
        provider_dir: Path,
        fault_polygons_inventory_df: pd.DataFrame,
    ) -> None:
        self._provider_id = provider_id
        self._provider_dir = provider_dir
        self._inventory_df = fault_polygons_inventory_df

    @staticmethod
    # pylint: disable=too-many-locals
    def write_backing_store(
        storage_dir: Path,
        storage_key: str,
        sim_fault_polygons: List[FaultPolygonsFileInfo],
    ) -> None:

        timer = PerfTimer()

        # All data for this provider will be stored inside a sub-directory
        # given by the storage key
        provider_dir = storage_dir / storage_key
        LOGGER.debug(f"Writing fault polygons backing store to: {provider_dir}")
        provider_dir.mkdir(parents=True, exist_ok=True)
        (provider_dir / REL_SIM_DIR).mkdir(parents=True, exist_ok=True)

        type_arr: List[FaultPolygonsType] = []
        real_arr: List[int] = []
        attribute_arr: List[str] = []
        name_arr: List[str] = []
        rel_path_arr: List[str] = []
        original_path_arr: List[str] = []

        for fault_polygons_info in sim_fault_polygons:
            rel_path_in_store = _compose_rel_sim_fault_polygons_path(
                real=fault_polygons_info.real,
                attribute=fault_polygons_info.attribute,
                name=fault_polygons_info.name,
                extension=Path(fault_polygons_info.path).suffix,
            )
            type_arr.append(FaultPolygonsType.SIMULATED)
            real_arr.append(fault_polygons_info.real)
            attribute_arr.append(fault_polygons_info.attribute)
            name_arr.append(fault_polygons_info.name)
            rel_path_arr.append(str(rel_path_in_store))
            original_path_arr.append(fault_polygons_info.path)

        LOGGER.debug(
            f"Copying {len(original_path_arr)} fault polygons into backing store..."
        )
        timer.lap_s()
        _copy_fault_polygons_into_provider_dir(
            original_path_arr, rel_path_arr, provider_dir
        )
        et_copy_s = timer.lap_s()

        fault_polygons_inventory_df = pd.DataFrame(
            {
                Col.TYPE: type_arr,
                Col.REAL: real_arr,
                Col.ATTRIBUTE: attribute_arr,
                Col.NAME: name_arr,
                Col.REL_PATH: rel_path_arr,
                Col.ORIGINAL_PATH: original_path_arr,
            }
        )

        parquet_file_name = provider_dir / "fault_polygons_inventory.parquet"
        fault_polygons_inventory_df.to_parquet(path=parquet_file_name)

        LOGGER.debug(
            f"Wrote fault polygons backing store in: {timer.elapsed_s():.2f}s ("
            f"copy={et_copy_s:.2f}s)"
        )

    @staticmethod
    def from_backing_store(
        storage_dir: Path,
        storage_key: str,
    ) -> Optional["ProviderImplFile"]:

        provider_dir = storage_dir / storage_key
        parquet_file_name = provider_dir / "fault_polygons_inventory.parquet"

        try:
            fault_polygons_inventory_df = pd.read_parquet(path=parquet_file_name)
            return ProviderImplFile(
                storage_key, provider_dir, fault_polygons_inventory_df
            )
        except FileNotFoundError:
            return None

    def provider_id(self) -> str:
        return self._provider_id

    def attributes(self) -> List[str]:
        return sorted(list(self._inventory_df[Col.ATTRIBUTE].unique()))

    def fault_polygons_names_for_attribute(
        self, fault_polygons_attribute: str
    ) -> List[str]:
        return sorted(
            list(
                self._inventory_df.loc[
                    self._inventory_df[Col.ATTRIBUTE] == fault_polygons_attribute
                ][Col.NAME].unique()
            )
        )

    def realizations(self) -> List[int]:
        unique_reals = self._inventory_df[Col.REAL].unique()

        # Sort and strip out any entries with real == -1
        return sorted([r for r in unique_reals if r >= 0])

    def get_fault_polygons(
        self,
        address: FaultPolygonsAddress,
    ) -> Optional[xtgeo.Polygons]:

        if isinstance(address, SimulatedFaultPolygonsAddress):
            return self._get_simulated_fault_polygons(address)

        raise TypeError("Unknown type of fault polygons address")

    def _get_simulated_fault_polygons(
        self, address: SimulatedFaultPolygonsAddress
    ) -> Optional[xtgeo.Polygons]:
        """Returns a Xtgeo fault polygons instance of a single realization fault polygons"""

        timer = PerfTimer()

        fault_polygons_fns: List[str] = self._locate_simulated_fault_polygons(
            attribute=address.attribute,
            name=address.name,
            realizations=[address.realization],
        )

        if len(fault_polygons_fns) == 0:
            LOGGER.warning(f"No simulated fault polygons found for {address}")
            return None
        if len(fault_polygons_fns) > 1:
            LOGGER.warning(
                f"Multiple simulated fault polygonss found for: {address}"
                "Returning first fault polygons."
            )

        fault_polygons = xtgeo.polygons_from_file(fault_polygons_fns[0])

        LOGGER.debug(f"Loaded simulated fault polygons in: {timer.elapsed_s():.2f}s")

        return fault_polygons

    def _locate_simulated_fault_polygons(
        self, attribute: str, name: str, realizations: List[int]
    ) -> List[str]:
        """Returns list of file names matching the specified filter criteria"""
        df = self._inventory_df.loc[
            self._inventory_df[Col.TYPE] == FaultPolygonsType.SIMULATED
        ]

        df = df.loc[
            (df[Col.ATTRIBUTE] == attribute)
            & (df[Col.NAME] == name)
            & (df[Col.REAL].isin(realizations))
        ]

        return [self._provider_dir / rel_path for rel_path in df[Col.REL_PATH]]


def _copy_fault_polygons_into_provider_dir(
    original_path_arr: List[str],
    rel_path_arr: List[str],
    provider_dir: Path,
) -> None:
    for src_path, dst_rel_path in zip(original_path_arr, rel_path_arr):
        # LOGGER.debug(f"copying fault polygons from: {src_path}")
        shutil.copyfile(src_path, provider_dir / dst_rel_path)

    # full_dst_path_arr = [storage_dir / dst_rel_path for dst_rel_path in store_path_arr]
    # with ProcessPoolExecutor() as executor:
    #     executor.map(shutil.copyfile, original_path_arr, full_dst_path_arr)


def _compose_rel_sim_fault_polygons_path(
    real: int,
    attribute: str,
    name: str,
    extension: str,
) -> Path:
    """Compose path to simulated fault polygons file, relative to provider's directory"""
    fname = f"{real}--{name}--{attribute}{extension}"
    return Path(REL_SIM_DIR) / fname



from typing import Any, Dict, List

import pandas as pd

from .._utils.colors import hex_to_rgba_str


class BarChart:
    """General bar chart class.

    Input:
    * series: pandas series with values and names in the index
    * n_rows: how many of the values to display
    * title
    * orientation
    """

    def __init__(self, series: pd.Series, n_rows: int, title: str, orientation: str):
        self.series = series.tail(n=n_rows)
        self.title = title
        self.orientation = orientation
        self._data = self.initial_data

    @property
    def data(self) -> List[Dict[str, Any]]:
        return self._data

    @property
    def figure(self) -> Dict[str, Any]:
        return {"data": self.data, "layout": self.layout}

    @property
    def initial_data(self) -> List[Dict[str, Any]]:
        return [
            {
                "x": self.series.values,
                "y": self.series.index,
                "orientation": self.orientation,
                "type": "bar",
                "marker": {},
            }
        ]

    @property
    def layout(self) -> Dict[str, Any]:
        return {
            "title": self.title,
            "margin": {
                "r": 20,
                "t": 60,
                "b": 20,
            },
            "barmode": "relative",
            "xaxis": {
                "range": [
                    self.series.values.min() - 0.1,
                    self.series.values.max() + 0.1,
                ]
            },
            "yaxis": {"automargin": True},
        }

    @property
    def first_y_value(self) -> str:
        return self.data[0]["y"][-1]

    def color_bars(
        self,
        selected_bar: str,
        color: str,
        opacity: float,
        color_selected: str = "#FF1243",
    ) -> None:
        """
        Set colors to the correlation plot bar,
        with separate color for the selected bar
        """
        self._data[0]["marker"] = {
            "color": [
                hex_to_rgba_str(color, opacity)
                if _bar != selected_bar
                else hex_to_rgba_str(color_selected, 0.8)
                for _bar in self._data[0]["y"]
            ],
            "line": {
                "color": [
                    color if _bar != selected_bar else color_selected
                    for _bar in self._data[0]["y"]
                ],
                "width": 1.2,
            },
        }



from typing import Dict, List, Optional, Tuple

import pandas as pd
import plotly.graph_objects as go


class PlotlyLinePlot:
    def __init__(
        self,
        xaxis_title: str = None,
        yaxis_title: str = None,
        ensemble_colors: Dict = None,
    ) -> None:
        self._ensemble_colors = ensemble_colors if ensemble_colors else {}
        self._realization_traces: List = []
        self._statistical_traces: List = []
        self._observation_traces: List = []
        self._layout: go.Layout = go.Layout(
            xaxis={"title": xaxis_title},
            yaxis={"title": yaxis_title},
            paper_bgcolor="rgba(0,0,0,0)",
            plot_bgcolor="rgba(0,0,0,0)",
        )

    def add_realization_traces(
        self,
        dframe: pd.DataFrame,
        x_column: str,
        y_column: str,
        color_column: Optional[str],
        highlight_reals: List = None,
        opacity: float = None,
        mode: str = "lines",
    ) -> None:
        """Renders line trace for each realization"""
        # If color parameter is given, normalize values for coloring
        if color_column is not None:
            dframe["VALUE_NORM"] = (
                dframe[color_column] - dframe[color_column].min()
            ) / (dframe[color_column].max() - dframe[color_column].min())
        highlight_reals = highlight_reals if highlight_reals else []
        for ensemble, ens_df in dframe.groupby("ENSEMBLE"):
            for real_no, (real, real_df) in enumerate(ens_df.groupby("REAL")):
                self._realization_traces.append(
                    {
                        "x": real_df[x_column],
                        "y": real_df[y_column],
                        "hovertemplate": (
                            f"Realization: {real}, Ensemble: {ensemble}"
                            f"<br>{color_column}: {real_df[color_column].unique()[0]}"
                        )
                        if color_column is not None
                        else f"Realization {real}, Ensemble: {ensemble}",
                        "name": ensemble,
                        "text": real,
                        "legendgroup": ensemble,
                        "marker": {
                            "color": "black"
                            if real in highlight_reals
                            else set_real_color(real_no=real, df_norm=dframe)
                            if color_column is not None
                            else self._ensemble_colors.get(
                                ensemble, "rgba(128,128,128,0.2)"
                            ),
                        },
                        "opacity": opacity,
                        "line": {"width": 3 if real in highlight_reals else 0.5},
                        "mode": mode,
                        "showlegend": real_no == 0 and color_column is None,
                    }
                )
        # Add a colorbar if parameter is used for coloring
        if color_column is not None:
            self._realization_traces.append(
                {
                    "x": [None],
                    "y": [None],
                    "mode": "markers",
                    "marker": {
                        "colorscale": [
                            ["0.0", "rgba(255,18,67, 1)"],
                            ["0.5", "rgba(220,220,220,1)"],
                            ["1.0", "rgba(62,208,62, 1)"],
                        ],
                        "cmin": dframe[color_column].min(),
                        "cmax": dframe[color_column].max(),
                        "colorbar": {
                            "thickness": 10,
                            "tickvals": [
                                dframe[color_column].min(),
                                dframe[color_column].max(),
                            ],
                            "tickmode": "array",
                        },
                        "showscale": True,
                    },
                    "hoverinfo": "none",
                    "showlegend": False,
                }
            )

    def add_statistical_lines(
        self,
        dframe: pd.DataFrame,
        x_column: str,
        y_column: str,
        traces: List,
        mode: str = "lines",
    ) -> None:
        for ensemble, ens_df in dframe.groupby("ENSEMBLE"):
            color = self._ensemble_colors.get(ensemble, "rgba(128,128,128,1)")
            if "Low/High" in traces:
                self._statistical_traces.append(
                    {
                        "line": {"dash": "dot", "width": 3},
                        "x": ens_df[x_column],
                        "y": ens_df[(y_column, "max")],
                        "hovertemplate": f"Calculation: {'mac'}, Ensemble: {ensemble}",
                        "name": ensemble,
                        "legendgroup": ensemble,
                        "showlegend": False,
                        "marker": {"color": color},
                        "mode": mode,
                    }
                )
            if "P10/P90" in traces:
                self._statistical_traces.append(
                    {
                        "line": {"dash": "dash"},
                        "x": ens_df[x_column],
                        "y": ens_df[(y_column, "high_p10")],
                        "hovertemplate": f"Calculation: {'high_p10'}, Ensemble: {ensemble}",
                        "name": ensemble,
                        "legendgroup": ensemble,
                        "showlegend": False,
                        "marker": {"color": color},
                        "mode": mode,
                    }
                )
            if "Mean" in traces:
                self._statistical_traces.append(
                    {
                        "x": ens_df[x_column],
                        "y": ens_df[(y_column, "mean")],
                        "hovertemplate": f"Calculation: {'mean'}, Ensemble: {ensemble}",
                        "name": ensemble,
                        "legendgroup": ensemble,
                        # "fill": "tonexty",
                        "marker": {"color": color},
                        "mode": mode,
                        "line": {"width": 3},
                    }
                )
            if "P10/P90" in traces:
                self._statistical_traces.append(
                    {
                        "line": {"dash": "dash"},
                        "x": ens_df[x_column],
                        "y": ens_df[(y_column, "low_p90")],
                        "hovertemplate": f"Calculation: {'low_p90'}, Ensemble: {ensemble}",
                        "name": ensemble,
                        "legendgroup": ensemble,
                        "showlegend": False,
                        # "fill": "tonexty",
                        "marker": {"color": color},
                        "mode": mode,
                    }
                )
            if "Low/High" in traces:
                self._statistical_traces.append(
                    {
                        "line": {"dash": "dot", "width": 1},
                        "x": ens_df[x_column],
                        "y": ens_df[(y_column, "min")],
                        "hovertemplate": f"Calculation: {'min'}, Ensemble: {ensemble}",
                        "name": ensemble,
                        "legendgroup": ensemble,
                        "showlegend": False,
                        "marker": {"color": color},
                        "mode": mode,
                    }
                )

    def add_observations(self, observations: list, x_value: str) -> None:
        for obs in observations:
            self._observation_traces.append(
                {
                    "x": [obs.get(x_value, [])],
                    "y": [obs.get("value", [])],
                    "marker": {"color": "black"},
                    "text": obs.get("comment", None),
                    "hoverinfo": "y+x+text",
                    "showlegend": False,
                    "error_y": {
                        "type": "data",
                        "array": [obs.get("error"), []],
                        "visible": True,
                    },
                }
            )

    def get_figure(self) -> Dict:
        traces = []
        if self._realization_traces:
            traces.extend(self._realization_traces)
        if self._statistical_traces:
            traces.extend(self._statistical_traces)
        if self._observation_traces:
            traces.extend(self._observation_traces)
        return dict(layout=self._layout, data=traces)


def set_real_color(df_norm: pd.DataFrame, real_no: str) -> str:
    """
    Return color for trace based on normalized parameter value.
    Midpoint for the colorscale is set on the average value
    """
    red = "rgba(255,18,67, 1)"
    mid_color = "rgba(220,220,220,1)"
    green = "rgba(62,208,62, 1)"
    df_norm = df_norm.reset_index(drop=True)
    mean = df_norm["VALUE_NORM"].mean()
    norm_value = df_norm.loc[df_norm["REAL"] == real_no].iloc[0]["VALUE_NORM"]
    if norm_value <= mean:
        intermed = norm_value / mean
        return find_intermediate_color(red, mid_color, intermed, colortype="rgba")
    if norm_value > mean:
        intermed = (norm_value - mean) / (1 - mean)
        return find_intermediate_color(mid_color, green, intermed, colortype="rgba")
    return "rgba(220,220,220, 0.8)"


def find_intermediate_color(
    lowcolor: str, highcolor: str, intermed: float, colortype: str = "tuple"
) -> str:
    """
    Returns the color at a given distance between two colors
    This function takes two color tuples, where each element is between 0
    and 1, along with a value 0 < intermed < 1 and returns a color that is
    intermed-percent from lowcolor to highcolor. If colortype is set to 'rgb',
    the function will automatically convert the rgb type to a tuple, find the
    intermediate color and return it as an rgb color.
    """

    if colortype == "rgba":
        # convert to tuple color, eg. (1, 0.45, 0.7)
        low = unlabel_rgba(lowcolor)
        high = unlabel_rgba(highcolor)

    diff_0 = float(high[0] - low[0])
    diff_1 = float(high[1] - low[1])
    diff_2 = float(high[2] - low[2])
    diff_3 = float(high[3] - low[3])

    inter_med_tuple = (
        low[0] + intermed * diff_0,
        low[1] + intermed * diff_1,
        low[2] + intermed * diff_2,
        low[3] + intermed * diff_3,
    )

    if colortype == "rgba":
        # back to an rgba string, e.g. rgba(30, 20, 10)
        inter_med_rgba = label_rgba(inter_med_tuple)
        return inter_med_rgba

    return str(inter_med_tuple)


def label_rgba(colors: Tuple) -> str:
    """
    Takes tuple (a, b, c, d) and returns an rgba color 'rgba(a, b, c, d)'
    """
    return f"rgba({colors[0]}, {colors[1]}, {colors[2]}, {colors[3]})"


def unlabel_rgba(colors: str) -> Tuple:
    """
    Takes rgba color(s) 'rgba(a, b, c, d)' and returns tuple(s) (a, b, c, d)
    This function takes either an 'rgba(a, b, c, d)' color or a list of
    such colors and returns the color tuples in tuple(s) (a, b, c, d)
    """
    str_vals = ""
    for index, _col in enumerate(colors):
        try:
            float(colors[index])
            str_vals = str_vals + colors[index]
        except ValueError:
            if colors[index] == "," or colors[index] == ".":
                str_vals = str_vals + colors[index]

    str_vals = str_vals + ","
    numbers = []
    str_num = ""
    for char in str_vals:
        if char != ",":
            str_num = str_num + char
        else:
            numbers.append(float(str_num))
            str_num = ""
    return tuple(numbers)



import datetime

import pytest

# pylint: disable=line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.datetime_utils import (
    from_str,
    to_str,
)


def test_from_str_success() -> None:
    assert from_str("2021-03-11") == datetime.datetime(2021, 3, 11)
    assert from_str("1956-08-26") == datetime.datetime(1956, 8, 26)


def test_from_str_assert() -> None:
    # Invalid datetime arguments (hour, minute, second, microsecond)
    invalid_dates = ["2021-03-11-23-55-11", "1996-05-26-23", "2001-08-11-11-43"]
    for _date in invalid_dates:
        with pytest.raises(ValueError) as err:
            from_str(_date)
        assert str(err.value) == f"unconverted data remains: {_date[10:]}"


def test_to_str_success() -> None:
    assert to_str(datetime.datetime(2021, 6, 13)) == "2021-06-13"
    assert to_str(datetime.datetime(2021, 12, 28)) == "2021-12-28"
    assert to_str(datetime.datetime(2021, 3, 7, 0)) == "2021-03-07"
    assert to_str(datetime.datetime(2021, 10, 22, 0, 0)) == "2021-10-22"
    assert to_str(datetime.datetime(2021, 1, 23, 0, 0, 0)) == "2021-01-23"
    assert to_str(datetime.datetime(2021, 12, 28, 0, 0, 0, 0)) == "2021-12-28"


def test_to_str_assert() -> None:
    # Invalid datetime arguments (hour, minute, second, microsecond)
    invalid_dates = [
        datetime.datetime(2021, 6, 13, 15, 32, 11, 43),
        datetime.datetime(2021, 6, 13, 5, 21, 45),
        datetime.datetime(2021, 6, 13, 23, 55),
        datetime.datetime(2021, 6, 13, 5),
    ]

    for _date in invalid_dates:
        with pytest.raises(ValueError) as err:
            to_str(_date)
        assert (
            str(err.value)
            == f"Invalid date resolution, expected no data for hour, minute, second"
            f" or microsecond for {str(_date)}"
        )



import datetime
from typing import Callable, Tuple, Union

import pandas as pd
from dash import ALL, Dash, Input, Output, State, callback_context, no_update
from dash._callback import NoUpdate
from dash.exceptions import PreventUpdate

from webviz_subsurface._figures import BarChart, ScatterPlot, TimeSeriesFigure
from webviz_subsurface._models import SurfaceLeafletModel
from webviz_subsurface._utils.dataframe_utils import (
    correlate_response_with_dataframe,
    merge_dataframes_on_realization,
)

from ..models import (
    PropertyStatisticsModel,
    ProviderTimeSeriesDataModel,
    SimulationTimeSeriesModel,
)
from ..utils.surface import surface_from_zone_prop


def property_response_controller(
    get_uuid: Callable,
    surface_table: pd.DataFrame,
    property_model: PropertyStatisticsModel,
    timeseries_model: Union[SimulationTimeSeriesModel, ProviderTimeSeriesDataModel],
    app: Dash,
) -> None:
    @app.callback(
        Output({"id": get_uuid("surface-view"), "tab": "response"}, "layers"),
        Output({"id": get_uuid("surface-name"), "tab": "response"}, "children"),
        Input(get_uuid("property-response-correlation-graph"), "clickData"),
        Input({"id": get_uuid("ensemble-selector"), "tab": "response"}, "value"),
        Input({"id": get_uuid("surface-type"), "tab": "response"}, "value"),
    )
    def _update_surface(
        clickdata: Union[None, dict], ensemble: str, stype: str
    ) -> Tuple[list, str]:
        if clickdata is not None:
            label = clickdata["points"][0]["y"]
            prop = label.split(" | ")[0]
            zone = label.split(" | ")[1]

            try:
                surface = surface_from_zone_prop(
                    surface_table,
                    zone=zone,
                    prop=prop,
                    ensemble=ensemble,
                    stype=stype,
                )
            except ValueError:  # Surface does not exist
                return (
                    no_update,
                    f"No surface found for {stype.capitalize()} for {prop}, {zone}",
                )
            surface_layer = SurfaceLeafletModel(surface, name="surface").layer
            return [surface_layer], f"{stype.capitalize()} for {prop}, {zone}"
        raise PreventUpdate

    @app.callback(
        Output(get_uuid("property-response-correlated-slider"), "min"),
        Output(get_uuid("property-response-correlated-slider"), "max"),
        Output(get_uuid("property-response-correlated-slider"), "step"),
        Output(get_uuid("property-response-correlated-slider"), "value"),
        Output(get_uuid("property-response-correlated-slider"), "marks"),
        Output(get_uuid("property-response-correlated-slider"), "disabled"),
        Input(get_uuid("property-response-correlated-filter"), "value"),
        Input({"id": get_uuid("ensemble-selector"), "tab": "response"}, "value"),
    )
    def _update_correlation_figure(
        label: str, ensemble: str
    ) -> Tuple[
        Union[float, NoUpdate],
        Union[float, NoUpdate],
        Union[float, NoUpdate],
        list,
        Union[dict, NoUpdate],
        bool,
    ]:
        if label is not None:
            values = property_model.filter_on_label(label=label, ensemble=ensemble)
            return (
                values.min(),
                values.max(),
                (values.max() - values.min()) / 100,
                [values.min(), values.max()],
                {
                    str(values.min()): {"label": f"{values.min():.2f}"},
                    str(values.max()): {"label": f"{values.max():.2f}"},
                },
                False,
            )

        return (
            no_update,
            no_update,
            no_update,
            [None, None],
            no_update,
            True,
        )

    @app.callback(
        Output(get_uuid("property-response-vector-graph"), "figure"),
        Output(get_uuid("property-response-correlation-graph"), "figure"),
        Output(get_uuid("property-response-scatter-graph"), "figure"),
        Input({"id": get_uuid("ensemble-selector"), "tab": "response"}, "value"),
        Input(get_uuid("property-response-vector-select"), "selectedNodes"),
        Input(get_uuid("property-response-vector-graph"), "clickData"),
        Input(get_uuid("property-response-correlation-graph"), "clickData"),
        Input(
            {
                "id": get_uuid("filter-selector"),
                "tab": "response",
                "selector": ALL,
            },
            "value",
        ),
        Input(get_uuid("property-response-correlated-slider"), "value"),
        State(get_uuid("property-response-vector-graph"), "figure"),
        State(get_uuid("property-response-correlated-filter"), "value"),
    )
    # pylint: disable=too-many-locals
    def _update_graphs(
        ensemble: str,
        vector: str,
        timeseries_clickdata: Union[None, dict],
        correlation_clickdata: Union[None, dict],
        selectors: list,
        corr_filter: list,
        figure: dict,
        label: str,
    ) -> Tuple[dict, dict, dict]:

        if (
            callback_context.triggered is None
            or callback_context.triggered[0]["prop_id"] == "."
            or vector is None
            or not vector
            or not all(filt for filt in selectors)
        ):
            raise PreventUpdate

        vector = vector[0]

        # Filter realizations if correlation filter is active
        real_filter = (
            property_model.filter_reals_on_label_range(ensemble, label, corr_filter)
            if corr_filter[0] is not None and corr_filter[1] is not None
            else None
        )

        # Get clicked data or last available date initially
        date = (
            timeseries_clickdata.get("points", [{}])[0].get(
                "x", timeseries_model.get_last_date(ensemble)
            )
            if timeseries_clickdata
            else timeseries_model.get_last_date(ensemble)
        )
        date = (
            datetime.datetime.strptime(date, "%Y-%m-%d")
            if isinstance(date, str)
            else date
        )

        # Get dataframe with vector and REAL
        vector_df = timeseries_model.get_vector_df(
            ensemble=ensemble, vectors=[vector], realizations=real_filter
        )
        if date not in vector_df["DATE"].values or vector not in vector_df:
            return {}, {}, {}

        # Get dataframe with properties per label and REAL
        prop_df = property_model.get_ensemble_properties(ensemble, selectors)
        prop_df = (
            prop_df[prop_df["REAL"].isin(real_filter)]
            if real_filter is not None
            else prop_df
        )

        merged_df = merge_dataframes_on_realization(
            dframe1=vector_df[vector_df["DATE"] == date], dframe2=prop_df
        )
        # Correlate properties against vector
        corrseries = correlate_response_with_dataframe(
            merged_df,
            response=vector,
            corrwith=[col for col in prop_df if col != "REAL"],
        )
        # Handle missing
        if corrseries.empty:
            return {}, {}, {}

        # Make correlation figure
        correlation_figure = BarChart(
            corrseries, n_rows=15, title=f"Correlations with {vector}", orientation="h"
        )

        # Get clicked correlation bar or largest bar initially
        selected_corr = (
            correlation_clickdata.get("points", [{}])[0].get("y")
            if correlation_clickdata
            else correlation_figure.first_y_value
        )

        # Update bar colors
        correlation_figure.color_bars(selected_corr, color="#007079", opacity=0.5)
        prop_df_norm = property_model.get_real_and_value_df(
            ensemble, series=selected_corr, normalize=True
        )
        figure = TimeSeriesFigure(
            dframe=merge_dataframes_on_realization(vector_df, prop_df_norm),
            visualization="realizations",
            vector=vector,
            ensemble=ensemble,
            dateline=date,
            historical_vector_df=timeseries_model.get_historical_vector_df(
                vector, ensemble
            ),
            color_col=selected_corr,
            line_shape_fallback=timeseries_model.line_shape_fallback,
        ).figure

        scatter_fig = (
            ScatterPlot(
                merged_df,
                response=vector,
                param=selected_corr,
                color="#007079",
                title=f"{vector} vs {selected_corr}",
                plot_trendline=True,
            ).figure
            if selected_corr in merged_df
            else {}
        )

        return figure, correlation_figure.figure, scatter_fig



from pathlib import Path
from typing import Dict, Optional

import pandas as pd

from webviz_subsurface._providers import (
    ColumnMetadata,
    EnsembleTableProvider,
    EnsembleTableProviderFactory,
)
from webviz_subsurface._providers.ensemble_table_provider import (
    EnsembleTableProviderImplArrow,
)


def _create_synthetic_table_provider(
    storage_dir: Path,
) -> EnsembleTableProvider:
    # fmt: off
    input_data = [
        ["REAL",  "A",   "B",  "STR" ],
        [     0,  1.0,  11.0,   "aa" ],
        [     0,  2.0,  12.0,   "bb" ],
        [     0,  3.0,  13.0,   "cc" ],
        [     1,  4.0,  14.0,   "dd" ],
        [     1,  5.0,  15.0,   "ee" ],
        [     1,  6.0,  16.0,   "ff" ],
        [     1,  7.0,  17.0,   "gg" ],
    ]
    # fmt: on

    input_df = pd.DataFrame(input_data[1:], columns=input_data[0])

    provider: Optional[EnsembleTableProvider]

    EnsembleTableProviderImplArrow.write_backing_store_from_ensemble_dataframe(
        storage_dir, "dummy_key", input_df
    )
    provider = EnsembleTableProviderImplArrow.from_backing_store(
        storage_dir, "dummy_key"
    )

    if not provider:
        raise ValueError("Failed to create EnsembleTableProvider")

    return provider


def test_synthetic_get_column_data(testdata_folder: Path) -> None:
    model = _create_synthetic_table_provider(testdata_folder)
    assert model.column_names() == ["A", "B", "STR"]
    assert model.realizations() == [0, 1]

    df = model.get_column_data(["A"])
    assert df.shape == (7, 2)
    assert df.columns.tolist() == ["REAL", "A"]

    df = model.get_column_data(["STR"], [1])
    assert df.shape == (4, 2)
    assert df.columns.tolist() == ["REAL", "STR"]

    assert model.column_metadata("REAL") is None


def test_create_from_aggregated_csv_file_smry_csv(
    testdata_folder: Path, tmp_path: Path
) -> None:
    factory = EnsembleTableProviderFactory(tmp_path, allow_storage_writes=True)
    provider = factory.create_from_ensemble_csv_file(
        testdata_folder / "reek_test_data" / "aggregated_data" / "smry.csv"
    )

    assert len(provider.column_names()) == 17
    assert provider.column_names()[0] == "DATE"
    assert provider.column_names()[16] == "YEARS"

    assert len(provider.realizations()) == 40

    valdf = provider.get_column_data(["YEARS"])
    assert len(valdf.columns) == 2
    assert valdf.columns[0] == "REAL"
    assert valdf.columns[1] == "YEARS"
    assert valdf["REAL"].nunique() == 40

    valdf = provider.get_column_data(["YEARS"], [0, 39, 10])
    assert len(valdf.columns) == 2
    assert valdf.columns[0] == "REAL"
    assert valdf.columns[1] == "YEARS"
    assert valdf["REAL"].nunique() == 3

    # No metadata in csv files
    meta: Optional[ColumnMetadata] = provider.column_metadata("FOPR")
    assert meta is None


def test_create_from_per_realization_csv_file(
    testdata_folder: Path, tmp_path: Path
) -> None:

    factory = EnsembleTableProviderFactory(tmp_path, allow_storage_writes=True)
    provider = factory.create_from_per_realization_csv_file(
        str(testdata_folder / "01_drogon_ahm/realization-*/iter-0"),
        "share/results/tables/rft.csv",
    )

    all_column_names = provider.column_names()
    # print(all_column_names)
    assert len(all_column_names) == 13

    assert len(provider.realizations()) == 100

    valdf = provider.get_column_data(["CONIDX"], [2])
    assert valdf.shape == (218, 2)
    assert valdf.columns[0] == "REAL"
    assert valdf.columns[1] == "CONIDX"
    assert valdf["REAL"].unique() == [2]
    assert valdf["CONIDX"].nunique() == 24
    assert sorted(valdf["CONIDX"].unique()) == list(range(1, 25))

    # No metadata in csv files
    meta: Optional[ColumnMetadata] = provider.column_metadata("CONIDX")
    assert meta is None


def test_create_from_per_realization_arrow_file(
    testdata_folder: Path, tmp_path: Path
) -> None:

    factory = EnsembleTableProviderFactory(tmp_path, allow_storage_writes=True)
    provider = factory.create_from_per_realization_arrow_file(
        str(testdata_folder / "01_drogon_ahm/realization-*/iter-0"),
        "share/results/unsmry/*arrow",
    )

    valdf = provider.get_column_data(provider.column_names())
    assert valdf.shape[0] == 25284
    assert "FOPT" in valdf.columns
    assert valdf["REAL"].nunique() == 100

    # Test metadata
    meta: Optional[ColumnMetadata] = provider.column_metadata("FOPR")
    assert meta is not None
    assert meta.unit == "SM3/DAY"


def test_create_from_per_realization_parameter_file(
    testdata_folder: Path, tmp_path: Path
) -> None:

    factory = EnsembleTableProviderFactory(tmp_path, allow_storage_writes=True)
    provider = factory.create_from_per_realization_parameter_file(
        str(testdata_folder / "01_drogon_ahm/realization-*/iter-0")
    )

    valdf = provider.get_column_data(provider.column_names())
    assert "GLOBVAR:FAULT_SEAL_SCALING" in valdf.columns
    assert valdf["REAL"].nunique() == 100

    # No metadata in parameter files
    meta: Optional[ColumnMetadata] = provider.column_metadata(
        "GLOBVAR:FAULT_SEAL_SCALING"
    )
    assert meta is None


def test_create_provider_set_from_aggregated_csv_file(tmp_path: Path) -> None:
    """This tests importing a csv file with an ensemble column with multiple
    ensembles. It will return a dictionary of providers, one for each ensemble.
    """
    factory = EnsembleTableProviderFactory(tmp_path, allow_storage_writes=True)
    provider_set: Dict[
        str, EnsembleTableProvider
    ] = factory.create_provider_set_from_aggregated_csv_file("tests/data/volumes.csv")
    assert set(provider_set.keys()) == {"iter-0", "iter-1"}

    for _, provider in provider_set.items():
        valdf = provider.get_column_data(provider.column_names())
        print(valdf)
        print(provider.column_names())
        assert set(valdf["REAL"].unique()) == {0, 1}
        assert {
            "ZONE",
            "REGION",
            "BULK_OIL",
            "PORE_OIL",
            "HCPV_OIL",
            "STOIIP_OIL",
            "SOURCE",
        }.issubset(set(provider.column_names()))

        # No metadata in csv files
        meta: Optional[ColumnMetadata] = provider.column_metadata("ZONE")
        assert meta is None



from datetime import datetime
from typing import List

import webviz_core_components as wcc
from dash import Input, Output, callback
from dash.development.base_component import Component
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from .._plugin_ids import PluginIds


class Filter(SettingsGroupABC):
    class Ids:
        # pylint: disable=too-few-public-methods
        ENSEMBLE_SELECTOR = "ensemble-selector"
        DATE_SELECTOR = "date-selector"
        PHASE_SELECTOR = "phase-selector"
        WELL_SELECTOR = "well-selector"
        COMBINE_WELL_AND_COLLECTION_AS = "combine-well-and-collection-as"
        WELL_COLLECTION_SELECTOR = "well-collection-selector"
        REALIZATION_SELECTOR = "realization-selector"

        FIG_LAYOUT_HEIGHT = "fig-layout-height"

    def __init__(
        self,
        ensemble_names: List[str],
        dates: List[datetime],
        phases: List[str],
        wells: List[str],
        realizations: List[int],
        all_well_collection_names: List[str],
    ) -> None:
        super().__init__("Filter")

        self.ensemble_names = ensemble_names
        self.dates = dates
        self.phases = phases
        self.wells = wells
        self.realizations = realizations
        self.all_well_collection_names = all_well_collection_names

    def layout(self) -> List[Component]:
        return [
            wcc.Checklist(
                label="Ensemble selector",
                id=self.register_component_unique_id(Filter.Ids.ENSEMBLE_SELECTOR),
                options=[{"label": ens, "value": ens} for ens in self.ensemble_names],
                value=self.ensemble_names[:2],
                vertical=False,
            ),
            wcc.SelectWithLabel(
                label="Date selector",
                id=self.register_component_unique_id(Filter.Ids.DATE_SELECTOR),
                options=[
                    {
                        "label": _date.strftime("%Y-%m-%d"),
                        "value": str(_date),
                    }
                    for _date in self.dates
                ],
                value=[str(self.dates[-1])],
                size=min([len(self.dates), 5]),
            ),
            wcc.SelectWithLabel(
                label="Phase selector",
                id=self.register_component_unique_id(Filter.Ids.PHASE_SELECTOR),
                options=[{"label": phase, "value": phase} for phase in self.phases],
                value=self.phases,
                size=min([len(self.phases), 3]),
            ),
            wcc.SelectWithLabel(
                label="Well selector",
                id=self.register_component_unique_id(Filter.Ids.WELL_SELECTOR),
                options=[{"label": well, "value": well} for well in self.wells],
                value=self.wells,
                size=min([len(self.wells), 9]),
            ),
            wcc.RadioItems(
                label="Combine wells and collections as",
                id=self.register_component_unique_id(
                    Filter.Ids.COMBINE_WELL_AND_COLLECTION_AS
                ),
                options=[
                    {
                        "label": "Intersection",
                        "value": "intersection",
                    },
                    {"label": "Union", "value": "union"},
                ],
                value="intersection",
            ),
            wcc.SelectWithLabel(
                label="Well collection selector",
                id=self.register_component_unique_id(
                    Filter.Ids.WELL_COLLECTION_SELECTOR
                ),
                options=[
                    {"label": collection, "value": collection}
                    for collection in self.all_well_collection_names
                ],
                value=self.all_well_collection_names,
                size=min([len(self.all_well_collection_names), 5]),
            ),
            wcc.SelectWithLabel(
                label="Realization selector",
                id=self.register_component_unique_id(Filter.Ids.REALIZATION_SELECTOR),
                options=[{"label": real, "value": real} for real in self.realizations],
                value=self.realizations,
                size=min([len(self.wells), 5]),
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_ENSEMBLES), "data"
            ),
            Input(
                self.component_unique_id(Filter.Ids.ENSEMBLE_SELECTOR).to_string(),
                "value",
            ),
        )
        def _set_ensembles(ensembles: List[str]) -> List[str]:
            return ensembles

        @callback(
            Output(self.get_store_unique_id(PluginIds.Stores.SELECTED_DATES), "data"),
            Input(
                self.component_unique_id(Filter.Ids.DATE_SELECTOR).to_string(),
                "value",
            ),
        )
        def _set_dates(dates: List[datetime]) -> List[datetime]:
            return dates

        @callback(
            Output(self.get_store_unique_id(PluginIds.Stores.SELECTED_PHASE), "data"),
            Input(
                self.component_unique_id(Filter.Ids.PHASE_SELECTOR).to_string(),
                "value",
            ),
        )
        def _set_phase(phase: List[str]) -> List[str]:
            return phase

        @callback(
            Output(self.get_store_unique_id(PluginIds.Stores.SELECTED_WELLS), "data"),
            Input(
                self.component_unique_id(Filter.Ids.WELL_SELECTOR).to_string(),
                "value",
            ),
        )
        def _set_wells(wells: List[str]) -> List[str]:
            return wells

        @callback(
            Output(
                self.get_store_unique_id(
                    PluginIds.Stores.SELECTED_COMBINE_WELLS_COLLECTION
                ),
                "data",
            ),
            Input(
                self.component_unique_id(
                    Filter.Ids.COMBINE_WELL_AND_COLLECTION_AS
                ).to_string(),
                "value",
            ),
        )
        def _set_combine_well_and_collection(combine_well_and_collection: str) -> str:
            return combine_well_and_collection

        @callback(
            Output(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_WELL_COLLECTIONS),
                "data",
            ),
            Input(
                self.component_unique_id(
                    Filter.Ids.WELL_COLLECTION_SELECTOR
                ).to_string(),
                "value",
            ),
        )
        def _set_well_collection_selector(well_collection: List[str]) -> List[str]:
            return well_collection

        @callback(
            Output(
                self.get_store_unique_id(PluginIds.Stores.SELECTED_REALIZATIONS), "data"
            ),
            Input(
                self.component_unique_id(Filter.Ids.REALIZATION_SELECTOR).to_string(),
                "value",
            ),
        )
        def _set_realizations(realizations: List[int]) -> List[int]:
            return realizations



import itertools
import logging
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple

import pandas as pd

from webviz_subsurface._utils.perf_timer import PerfTimer

from ...._models import StratigraphyModel, WellAttributesModel
from ...._providers import EnsembleTableProvider

LOGGER = logging.getLogger(__name__)


class WellCompletionDataModel:
    def __init__(
        self,
        ensemble_path: str,
        wellcompletion_provider: EnsembleTableProvider,
        stratigraphy_model: StratigraphyModel,
        well_attributes_model: WellAttributesModel,
        theme_colors: List[str],
    ) -> None:
        self._ensemble_path = ensemble_path
        self._theme_colors = theme_colors
        self._stratigraphy_model = stratigraphy_model
        self._stratigraphy = self._stratigraphy_model.data
        self._well_attributes_model = well_attributes_model
        self._well_attributes = self._well_attributes_model.data

        self._wellcompletion_df = wellcompletion_provider.get_column_data(
            column_names=wellcompletion_provider.column_names()
        )
        self._zones = self._wellcompletion_df["ZONE"].unique()
        self._realizations = sorted(self._wellcompletion_df["REAL"].unique())
        self._datemap = {
            dte: i
            for i, dte in enumerate(sorted(self._wellcompletion_df["DATE"].unique()))
        }
        self._wellcompletion_df["TIMESTEP"] = self._wellcompletion_df["DATE"].map(
            self._datemap
        )
        kh_metadata = wellcompletion_provider.column_metadata("KH")
        self._kh_unit = (
            kh_metadata.unit
            if kh_metadata is not None and kh_metadata.unit is not None
            else ""
        )
        self._kh_decimal_places = 2

    @property
    def webviz_store(self) -> List[Tuple[Callable, List[Dict]]]:
        return [
            self._well_attributes_model.webviz_store,
            self._stratigraphy_model.webviz_store,
        ]

    @property
    def realizations(self) -> List[int]:
        return self._realizations

    def create_ensemble_dataset(
        self, realization: Optional[int] = None
    ) -> Dict[str, Any]:
        """Creates the input data set for the WellCompletions component.

        Returns a dictionary on a given format specified here:
        https://github.com/equinor/webviz-subsurface-components/blob/master/react/src/lib/inputSchema/wellCompletions.json

        if realization is not None, the input dataframe is filtered on realizations.
        """
        timer = PerfTimer()

        df = self._wellcompletion_df
        if realization is not None:
            df = df[df["REAL"] == realization]

        dataset = {
            "version": "1.1.0",
            "units": {
                "kh": {"unit": self._kh_unit, "decimalPlaces": self._kh_decimal_places}
            },
            "stratigraphy": self._extract_stratigraphy(),
            "timeSteps": [
                pd.to_datetime(str(dte)).strftime("%Y-%m-%d")
                for dte in self._datemap.keys()
            ],
            "wells": self._extract_wells(df),
        }

        LOGGER.info(f"WellCompletion dataset created in {timer.elapsed_s():.2f}s")

        return dataset

    def _extract_wells(self, wellcompletion_df: pd.DataFrame) -> List[Dict[str, Any]]:
        """Generates the wells part of the input to the WellCompletions component."""
        well_list = []
        no_real = wellcompletion_df["REAL"].nunique()
        for well_name, well_group in wellcompletion_df.groupby("WELL"):
            well_data = _extract_well(well_group, well_name, no_real)
            well_data["attributes"] = (
                self._well_attributes[well_name]
                if well_name in self._well_attributes
                else {}
            )
            well_list.append(well_data)
        return well_list

    def _extract_stratigraphy(self) -> List[Dict[str, Any]]:
        """Returns the stratigraphy part of the input to the WellCompletions component."""
        color_iterator = itertools.cycle(self._theme_colors)

        # If no stratigraphy file is found then the stratigraphy is
        # created from the unique zones in the wellcompletiondata input.
        # They will then probably not come in the correct order.
        if self._stratigraphy is None:
            return [
                {
                    "name": zone,
                    "color": next(color_iterator),
                }
                for zone in self._zones
            ]

        # If stratigraphy is not None the following is done:
        stratigraphy, remaining_valid_zones = _filter_valid_nodes(
            self._stratigraphy, self._zones
        )

        if remaining_valid_zones:
            raise ValueError(
                "The following zones are defined in the well completion data, "
                f"but not in the stratigraphy: {remaining_valid_zones}"
            )

        return _add_colors_to_stratigraphy(stratigraphy, color_iterator)


def _add_colors_to_stratigraphy(
    stratigraphy: List[Dict[str, Any]],
    color_iterator: Iterator,
    zone_color_mapping: Optional[Dict[str, str]] = None,
) -> List[Dict[str, Any]]:
    """Add colors to the stratigraphy tree. The function will recursively parse the tree.

    There are tree sources of color:
    1. The color is given in the stratigraphy list, in which case nothing is done to the node
    2. The color is the optional the zone->color map
    3. If none of the above applies, the color will be taken from the theme color iterable for \
    the leaves. For other levels, a dummy color grey is used
    """
    for zonedict in stratigraphy:
        if "color" not in zonedict:
            if (
                zone_color_mapping is not None
                and zonedict["name"] in zone_color_mapping
            ):
                zonedict["color"] = zone_color_mapping[zonedict["name"]]
            elif "subzones" not in zonedict:
                zonedict["color"] = next(
                    color_iterator
                )  # theme colors only applied on leaves
            else:
                zonedict["color"] = "#808080"  # grey
        if "subzones" in zonedict:
            zonedict["subzones"] = _add_colors_to_stratigraphy(
                zonedict["subzones"],
                color_iterator,
                zone_color_mapping=zone_color_mapping,
            )
    return stratigraphy


def _extract_well(
    well_group: pd.DataFrame, well_name: str, no_real: int
) -> Dict[str, Any]:
    """Extract completion events and kh values for a single well"""
    well_dict: Dict[str, Any] = {}
    well_dict["name"] = well_name

    completions: Dict[str, Dict[str, List[Any]]] = {}
    for (zone, timestep), group_df in well_group.groupby(["ZONE", "TIMESTEP"]):
        data = group_df["OP/SH"].value_counts()
        if zone not in completions:
            completions[zone] = {
                "t": [],
                "open": [],
                "shut": [],
                "khMean": [],
                "khMin": [],
                "khMax": [],
            }
        zonedict = completions[zone]
        zonedict["t"].append(int(timestep))
        zonedict["open"].append(float(data["OPEN"] / no_real if "OPEN" in data else 0))
        zonedict["shut"].append(float(data["SHUT"] / no_real if "SHUT" in data else 0))
        zonedict["khMean"].append(round(float(group_df["KH"].mean()), 2))
        zonedict["khMin"].append(round(float(group_df["KH"].min()), 2))
        zonedict["khMax"].append(round(float(group_df["KH"].max()), 2))

    well_dict["completions"] = completions
    return well_dict


def _filter_valid_nodes(
    stratigraphy: List[Dict[str, Any]], valid_zone_names: List[str]
) -> Tuple[List, List]:
    """Returns the stratigraphy tree with only valid nodes.
    A node is considered valid if it self or one of it's subzones are in the
    valid zone names list (passed from the lyr file)

    The function recursively parses the tree to add valid nodes.
    """
    output = []
    remaining_valid_zones = valid_zone_names
    for zonedict in stratigraphy:
        if "subzones" in zonedict:
            zonedict["subzones"], remaining_valid_zones = _filter_valid_nodes(
                zonedict["subzones"], remaining_valid_zones
            )
        if zonedict["name"] in remaining_valid_zones:
            if "subzones" in zonedict and not zonedict["subzones"]:
                zonedict.pop("subzones")
            output.append(zonedict)
            remaining_valid_zones = [
                zone for zone in remaining_valid_zones if zone != zonedict["name"]
            ]  # remove zone name from valid zones if it is found in the stratigraphy
        elif "subzones" in zonedict and zonedict["subzones"]:
            output.append(zonedict)

    return output, remaining_valid_zones



from typing import List, Optional

import webviz_core_components as wcc
from dash import html
from webviz_config import WebvizConfigTheme

from webviz_subsurface._models import InplaceVolumesModel


def selections_layout(
    uuid: str,
    volumemodel: InplaceVolumesModel,
    theme: WebvizConfigTheme,
    tab: str,
) -> html.Div:
    selectors = "/".join(
        [
            x.lower()
            for x in ["ZONE", "REGION", "FACIES", "FIPNUM", "SET"]
            if x in volumemodel.selectors
        ]
    )
    return html.Div(
        children=[
            html.Div(
                style={"margin-bottom": "20px"},
                children=[
                    button(uuid=uuid, title="Custom plotting", page_id="custom"),
                    button(uuid=uuid, title=f"Plots per {selectors}", page_id="per_zr"),
                    button(
                        uuid=uuid, title="Convergence plot mean/p10/p90", page_id="conv"
                    ),
                ],
            ),
            plot_selections_layout(uuid, volumemodel, tab),
            settings_layout(volumemodel, uuid, theme, tab),
        ]
    )


def button(
    uuid: str,
    title: str,
    page_id: str,
) -> html.Button:
    return html.Button(
        title,
        className="webviz-inplace-vol-btn",
        id={"id": uuid, "button": page_id},
    )


def plot_selections_layout(
    uuid: str, volumemodel: InplaceVolumesModel, tab: str
) -> wcc.Selectors:
    return wcc.Selectors(
        label="PLOT CONTROLS",
        open_details=True,
        children=plot_selector_dropdowns(uuid=uuid, volumemodel=volumemodel, tab=tab)
        + [
            html.Div(
                style={"margin-top": "10px"},
                children=wcc.RadioItems(
                    label="Visualization below plot:",
                    id={"id": uuid, "tab": tab, "selector": "bottom_viz"},
                    options=[
                        {"label": "Table", "value": "table"},
                        {"label": "None", "value": "none"},
                    ],
                    vertical=False,
                    value="table",
                ),
            ),
        ],
    )


def table_selections_layout(
    uuid: str, volumemodel: InplaceVolumesModel, tab: str
) -> wcc.Selectors:
    responses = volumemodel.volume_columns + volumemodel.property_columns
    return wcc.Selectors(
        label="TABLE CONTROLS",
        open_details=True,
        children=[
            wcc.Dropdown(
                label="Table type",
                id={"id": uuid, "tab": tab, "selector": "Table type"},
                options=[
                    {"label": elm, "value": elm}
                    for elm in ["Statistics table", "Mean table"]
                ],
                value="Mean table",
                clearable=False,
            ),
            wcc.Dropdown(
                label="Group by",
                id={"id": uuid, "tab": tab, "selector": "Group by"},
                options=[{"label": elm, "value": elm} for elm in volumemodel.selectors],
                value=None,
                multi=True,
                clearable=False,
            ),
            wcc.SelectWithLabel(
                label="Responses",
                id={"id": uuid, "tab": tab, "selector": "table_responses"},
                options=[{"label": i, "value": i} for i in responses],
                value=responses,
                size=min(20, len(responses)),
            ),
        ],
    )


def plot_selector_dropdowns(
    uuid: str, volumemodel: InplaceVolumesModel, tab: str
) -> List[html.Div]:
    """Makes dropdowns for each selector"""

    dropdowns: List[html.Div] = []
    value: Optional[str] = None

    for selector in [
        "Plot type",
        "X Response",
        "Y Response",
        "Subplots",
        "Color by",
    ]:
        if selector == "Plot type":
            elements = ["histogram", "scatter", "distribution", "box", "bar"]
            value = elements[0]
        if selector == "X Response":
            elements = volumemodel.responses
            value = elements[0]
        if selector == "Y Response":
            elements = volumemodel.responses
            value = None
        if selector == "Subplots":
            elements = [x for x in volumemodel.selectors if x != "REAL"]
            value = None
        if selector == "Color by":
            elements = [x for x in volumemodel.selectors if x != "REAL"]
            value = "ENSEMBLE"

        dropdowns.append(
            wcc.Dropdown(
                label=selector,
                id={"id": uuid, "tab": tab, "selector": selector},
                options=[{"label": elm, "value": elm} for elm in elements],
                value=value,
                clearable=selector in ["Subplots", "Color by", "Y Response"],
                disabled=selector == "Y Response",
            )
        )
    return dropdowns


def settings_layout(
    volumemodel: InplaceVolumesModel, uuid: str, theme: WebvizConfigTheme, tab: str
) -> wcc.Selectors:

    theme_colors = theme.plotly_theme.get("layout", {}).get("colorway", [])
    return wcc.Selectors(
        label="⚙️ SETTINGS",
        open_details=False,
        children=[
            remove_fluid_annotation(volumemodel, uuid=uuid, tab=tab),
            subplot_xaxis_range(uuid=uuid, tab=tab),
            histogram_options(uuid=uuid, tab=tab),
            html.Span("Colors", style={"font-weight": "bold"}),
            wcc.ColorScales(
                id={"id": uuid, "tab": tab, "settings": "Colorscale"},
                colorscale=theme_colors,
                fixSwatches=True,
                nSwatches=12,
            ),
        ],
    )


def subplot_xaxis_range(uuid: str, tab: str) -> html.Div:
    axis_matches_layout = []
    for axis in ["X axis", "Y axis"]:
        axis_matches_layout.append(
            html.Div(
                children=wcc.Checklist(
                    id={"id": uuid, "tab": tab, "selector": f"{axis} matches"},
                    options=[{"label": f"Equal {axis} range", "value": "Equal"}],
                    value=["Equal"],
                )
            )
        )
    return html.Div(
        children=[
            html.Span("Subplot options:", style={"font-weight": "bold"}),
            html.Div(style={"margin-bottom": "10px"}, children=axis_matches_layout),
        ]
    )


def table_sync_option(uuid: str, tab: str) -> html.Div:
    return html.Div(
        style={"margin-bottom": "10px"},
        children=wcc.Checklist(
            id={"id": uuid, "tab": tab, "selector": "sync_table"},
            options=[{"label": "Sync table with plot", "value": "Sync"}],
            value=["Sync"],
        ),
    )


def remove_fluid_annotation(
    volumemodel: InplaceVolumesModel, uuid: str, tab: str
) -> html.Div:
    return html.Div(
        style={
            "margin-bottom": "10px",
            "display": "none" if volumemodel.volume_type == "dynamic" else "block",
        },
        children=wcc.Checklist(
            id={"id": uuid, "tab": tab, "selector": "Fluid annotation"},
            options=[{"label": "Show fluid annotation", "value": "Show"}],
            value=["Show"] if volumemodel.volume_type != "dynamic" else [],
        ),
    )


def histogram_options(uuid: str, tab: str) -> html.Div:
    return html.Div(
        children=[
            wcc.RadioItems(
                label="Barmode:",
                id={"id": uuid, "tab": tab, "selector": "barmode"},
                options=[
                    {"label": "overlay", "value": "overlay"},
                    {"label": "group", "value": "group"},
                    {"label": "stack", "value": "stack"},
                ],
                labelStyle={"display": "inline-flex", "margin-right": "5px"},
                value="overlay",
            ),
            wcc.Slider(
                label="Histogram bins:",
                id={"id": uuid, "tab": tab, "selector": "hist_bins"},
                value=15,
                step=1,
                marks={1: 1, 10: 10, 20: 20, 30: 30},
                min=1,
                max=30,
            ),
        ]
    )



import datetime

import pandas as pd
import pytest
from pandas._testing import assert_frame_equal

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.dataframe_utils import make_date_column_datetime_object

# pylint: disable=line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.from_timeseries_cumulatives import (
    calculate_from_resampled_cumulative_vectors_df,
    create_per_day_vector_name,
    create_per_interval_vector_name,
    datetime_to_intervalstr,
    get_cumulative_vector_name,
    is_per_interval_or_per_day_vector,
)

# *******************************************************************
#####################################################################
#
# CONFIGURE TESTDATA
#
#####################################################################
# *******************************************************************

# fmt: off
# Monthly frequency - rate per day implies divide on days in month
INPUT_WEEKLY_DF = pd.DataFrame(
    columns=["DATE", "REAL", "A", "B"],
    data=[
        [datetime.datetime(2021, 1, 1),  1, 50.0,   250.0 ],
        [datetime.datetime(2021, 1, 8),  1, 100.0,  500.0 ],
        [datetime.datetime(2021, 1, 15), 1, 150.0,  750.0 ],
        [datetime.datetime(2021, 1, 1),  2, 300.0,  350.0 ],
        [datetime.datetime(2021, 1, 8),  2, 400.0,  700.0 ],
        [datetime.datetime(2021, 1, 15), 2, 500.0,  1050.0],
        [datetime.datetime(2021, 1, 1),  4, 1000.0, 450.0 ],
        [datetime.datetime(2021, 1, 8),  4, 1200.0, 900.0 ],
        [datetime.datetime(2021, 1, 15), 4, 1400.0, 1350.0],
    ],
)
EXPECTED_PER_INTVL_WEEKLY_DF = pd.DataFrame(
    columns=["DATE", "REAL", "PER_INTVL_A", "PER_INTVL_B"],
    data=[
        [datetime.datetime(2021, 1, 1),  1, 50.0,  250.0],
        [datetime.datetime(2021, 1, 8),  1, 50.0,  250.0],
        [datetime.datetime(2021, 1, 15), 1, 0.0,   0.0  ],
        [datetime.datetime(2021, 1, 1),  2, 100.0, 350.0],
        [datetime.datetime(2021, 1, 8),  2, 100.0, 350.0],
        [datetime.datetime(2021, 1, 15), 2, 0.0,   0.0  ],
        [datetime.datetime(2021, 1, 1),  4, 200.0, 450.0],
        [datetime.datetime(2021, 1, 8),  4, 200.0, 450.0],
        [datetime.datetime(2021, 1, 15), 4, 0.0,   0.0  ],
    ],
)
EXPECTED_PER_DAY_WEEKLY_DF = pd.DataFrame(
    columns=["DATE", "REAL", "PER_DAY_A", "PER_DAY_B"],
    data=[
        [datetime.datetime(2021, 1, 1),  1, 50.0/7.0,  250.0/7.0],
        [datetime.datetime(2021, 1, 8),  1, 50.0/7.0,  250.0/7.0],
        [datetime.datetime(2021, 1, 15), 1, 0.0,       0.0      ],
        [datetime.datetime(2021, 1, 1),  2, 100.0/7.0, 350.0/7.0],
        [datetime.datetime(2021, 1, 8),  2, 100.0/7.0, 350.0/7.0],
        [datetime.datetime(2021, 1, 15), 2, 0.0,       0.0      ],
        [datetime.datetime(2021, 1, 1),  4, 200.0/7.0, 450.0/7.0],
        [datetime.datetime(2021, 1, 8),  4, 200.0/7.0, 450.0/7.0],
        [datetime.datetime(2021, 1, 15), 4, 0.0,       0.0      ],
    ],
)
# Convert date columns to datetime.datetime
make_date_column_datetime_object(INPUT_WEEKLY_DF)
make_date_column_datetime_object(EXPECTED_PER_INTVL_WEEKLY_DF)
make_date_column_datetime_object(EXPECTED_PER_DAY_WEEKLY_DF)

# Monthly frequency - rate per day implies divide on days in month
INPUT_MONTHLY_DF = pd.DataFrame(
    columns=["DATE", "REAL", "A", "B"],
    data=[
        [datetime.datetime(2021, 1, 1), 1, 50.0,   250.0 ],
        [datetime.datetime(2021, 2, 1), 1, 100.0,  500.0 ],
        [datetime.datetime(2021, 3, 1), 1, 150.0,  750.0 ],
        [datetime.datetime(2021, 1, 1), 2, 300.0,  350.0 ],
        [datetime.datetime(2021, 2, 1), 2, 400.0,  700.0 ],
        [datetime.datetime(2021, 3, 1), 2, 500.0,  1050.0],
        [datetime.datetime(2021, 1, 1), 4, 1000.0, 450.0 ],
        [datetime.datetime(2021, 2, 1), 4, 1200.0, 900.0 ],
        [datetime.datetime(2021, 3, 1), 4, 1400.0, 1350.0],
    ],
)
EXPECTED_PER_INTVL_MONTHLY_DF = pd.DataFrame(
    columns=["DATE", "REAL", "PER_INTVL_A", "PER_INTVL_B"],
    data=[
        [datetime.datetime(2021, 1, 1), 1, 50.0,  250.0],
        [datetime.datetime(2021, 2, 1), 1, 50.0,  250.0],
        [datetime.datetime(2021, 3, 1), 1, 0.0,   0.0  ],
        [datetime.datetime(2021, 1, 1), 2, 100.0, 350.0],
        [datetime.datetime(2021, 2, 1), 2, 100.0, 350.0],
        [datetime.datetime(2021, 3, 1), 2, 0.0,   0.0  ],
        [datetime.datetime(2021, 1, 1), 4, 200.0, 450.0],
        [datetime.datetime(2021, 2, 1), 4, 200.0, 450.0],
        [datetime.datetime(2021, 3, 1), 4, 0.0,   0.0  ],
    ],
)
EXPECTED_PER_DAY_MONTHLY_DF = pd.DataFrame(
    columns=["DATE", "REAL", "PER_DAY_A", "PER_DAY_B"],
    data=[
        [datetime.datetime(2021, 1, 1), 1, 50.0/31.0,  250.0/31.0],
        [datetime.datetime(2021, 2, 1), 1, 50.0/28.0,  250.0/28.0],
        [datetime.datetime(2021, 3, 1), 1, 0.0,        0.0       ],
        [datetime.datetime(2021, 1, 1), 2, 100.0/31.0, 350.0/31.0],
        [datetime.datetime(2021, 2, 1), 2, 100.0/28.0, 350.0/28.0],
        [datetime.datetime(2021, 3, 1), 2, 0.0,        0.0       ],
        [datetime.datetime(2021, 1, 1), 4, 200.0/31.0, 450.0/31.0],
        [datetime.datetime(2021, 2, 1), 4, 200.0/28.0, 450.0/28.0],
        [datetime.datetime(2021, 3, 1), 4, 0.0,        0.0       ],
    ],
)
# Convert date columns to datetime.datetime
make_date_column_datetime_object(INPUT_MONTHLY_DF)
make_date_column_datetime_object(EXPECTED_PER_INTVL_MONTHLY_DF)
make_date_column_datetime_object(EXPECTED_PER_DAY_MONTHLY_DF)

# Yearly frequency - rate per day implies divide on days in year
INPUT_YEARLY_DF = pd.DataFrame(
    columns=["DATE", "REAL", "A", "B"],
    data=[
        [datetime.datetime(2021, 1, 1), 1, 50.0,   250.0 ],
        [datetime.datetime(2022, 1, 1), 1, 100.0,  500.0 ],
        [datetime.datetime(2023, 1, 1), 1, 150.0,  750.0 ],
        [datetime.datetime(2021, 1, 1), 2, 300.0,  350.0 ],
        [datetime.datetime(2022, 1, 1), 2, 400.0,  700.0 ],
        [datetime.datetime(2023, 1, 1), 2, 500.0,  1050.0],
        [datetime.datetime(2021, 1, 1), 4, 1000.0, 450.0 ],
        [datetime.datetime(2022, 1, 1), 4, 1200.0, 900.0 ],
        [datetime.datetime(2023, 1, 1), 4, 1400.0, 1350.0],
    ],
)
EXPECTED_PER_INTVL_YEARLY_DF = pd.DataFrame(
    columns=["DATE", "REAL", "PER_INTVL_A", "PER_INTVL_B"],
    data=[
        [datetime.datetime(2021, 1, 1), 1, 50.0,  250.0],
        [datetime.datetime(2022, 1, 1), 1, 50.0,  250.0],
        [datetime.datetime(2023, 1, 1), 1, 0.0,   0.0  ],
        [datetime.datetime(2021, 1, 1), 2, 100.0, 350.0],
        [datetime.datetime(2022, 1, 1), 2, 100.0, 350.0],
        [datetime.datetime(2023, 1, 1), 2, 0.0,   0.0  ],
        [datetime.datetime(2021, 1, 1), 4, 200.0, 450.0],
        [datetime.datetime(2022, 1, 1), 4, 200.0, 450.0],
        [datetime.datetime(2023, 1, 1), 4, 0.0,   0.0  ],
    ],
)
EXPECTED_PER_DAY_YEARLY_DF = pd.DataFrame(
    columns=["DATE", "REAL", "PER_DAY_A", "PER_DAY_B"],
    data=[
        [datetime.datetime(2021, 1, 1), 1, 50.0/365.0,  250.0/365.0],
        [datetime.datetime(2022, 1, 1), 1, 50.0/365.0,  250.0/365.0],
        [datetime.datetime(2023, 1, 1), 1, 0.0,         0.0        ],
        [datetime.datetime(2021, 1, 1), 2, 100.0/365.0, 350.0/365.0],
        [datetime.datetime(2022, 1, 1), 2, 100.0/365.0, 350.0/365.0],
        [datetime.datetime(2023, 1, 1), 2, 0.0,         0.0        ],
        [datetime.datetime(2021, 1, 1), 4, 200.0/365.0, 450.0/365.0],
        [datetime.datetime(2022, 1, 1), 4, 200.0/365.0, 450.0/365.0],
        [datetime.datetime(2023, 1, 1), 4, 0.0,         0.0        ],
    ],
)
# Convert date columns to datetime.datetime
make_date_column_datetime_object(INPUT_YEARLY_DF)
make_date_column_datetime_object(EXPECTED_PER_INTVL_YEARLY_DF)
make_date_column_datetime_object(EXPECTED_PER_DAY_YEARLY_DF)


# Monthly frequency after year 2262 - rate per day implies divide on days in month
AFTER_2262_MONTHLY_DATES = pd.Series(
    [
        datetime.datetime(2265, 1, 1),
        datetime.datetime(2265, 2, 1),
        datetime.datetime(2265, 3, 1),
        datetime.datetime(2265, 1, 1),
        datetime.datetime(2265, 2, 1),
        datetime.datetime(2265, 3, 1),
        datetime.datetime(2265, 1, 1),
        datetime.datetime(2265, 2, 1),
        datetime.datetime(2265, 3, 1),
    ]
)
# NOTE: datetime.datetime after year 2262 is not converted to pd.Timestamp!
INPUT_MONTHLY_AFTER_2262_DF = INPUT_MONTHLY_DF.copy()
INPUT_MONTHLY_AFTER_2262_DF["DATE"] = AFTER_2262_MONTHLY_DATES
EXPECTED_PER_INTVL_MONTHLY_AFTER_2262_DF = EXPECTED_PER_INTVL_MONTHLY_DF.copy()
EXPECTED_PER_INTVL_MONTHLY_AFTER_2262_DF["DATE"] = AFTER_2262_MONTHLY_DATES
EXPECTED_PER_DAY_MONTHLY_AFTER_2262_DF = EXPECTED_PER_DAY_MONTHLY_DF.copy()
EXPECTED_PER_DAY_MONTHLY_AFTER_2262_DF["DATE"] = AFTER_2262_MONTHLY_DATES

# fmt: on


# *******************************************************************
#####################################################################
#
# UNIT TESTS
#
#####################################################################
# *******************************************************************

TEST_CASES = [
    pytest.param(
        INPUT_WEEKLY_DF, EXPECTED_PER_INTVL_WEEKLY_DF, EXPECTED_PER_DAY_WEEKLY_DF
    ),
    pytest.param(
        INPUT_MONTHLY_DF, EXPECTED_PER_INTVL_MONTHLY_DF, EXPECTED_PER_DAY_MONTHLY_DF
    ),
    pytest.param(
        INPUT_YEARLY_DF, EXPECTED_PER_INTVL_YEARLY_DF, EXPECTED_PER_DAY_YEARLY_DF
    ),
    pytest.param(
        INPUT_MONTHLY_AFTER_2262_DF,
        EXPECTED_PER_INTVL_MONTHLY_AFTER_2262_DF,
        EXPECTED_PER_DAY_MONTHLY_AFTER_2262_DF,
    ),
]


@pytest.mark.parametrize(
    "input_df, expected_per_intvl_df, expected_per_day_df", TEST_CASES
)
def test_calculate_from_resampled_cumulative_vectors_df(
    input_df: pd.DataFrame,
    expected_per_intvl_df: pd.DataFrame,
    expected_per_day_df: pd.DataFrame,
) -> None:
    # PER_INTVL_ due to as_rate_per_day = False
    calculated_per_intvl_df = calculate_from_resampled_cumulative_vectors_df(
        input_df, False
    )

    # PER_DAY_ due to as_rate_per_day = True
    calculated_per_day_df = calculate_from_resampled_cumulative_vectors_df(
        input_df, True
    )

    assert_frame_equal(expected_per_intvl_df, calculated_per_intvl_df)
    assert_frame_equal(expected_per_day_df, calculated_per_day_df)


def test_calculate_from_resampled_cumulative_vectors_df_invalid_input() -> None:
    """Test assert check assert_date_column_is_datetime_object() in
    webviz_subsurface._utils.dataframe_utils.py
    """
    # fmt: off
    input_df = pd.DataFrame(
        columns=["DATE", "REAL", "A", "B"],
        data=[
            [pd.Timestamp(2020, 1, 1), 1, 50.0,   250.0 ],
            [pd.Timestamp(2020, 2, 1), 1, 100.0,  500.0 ],
            [pd.Timestamp(2020, 3, 1), 1, 150.0,  750.0 ],
            [pd.Timestamp(2020, 4, 1), 1, 200.0,  1000.0],
            [pd.Timestamp(2020, 5, 1), 1, 250.0,  1250.0],
        ],
    )
    # fmt: on

    with pytest.raises(ValueError) as err:
        calculate_from_resampled_cumulative_vectors_df(input_df, True)
    assert (
        str(err.value)
        == '"DATE"-column in dataframe is not on datetime.datetime format!'
    )


def test_is_per_interval_or_per_day_vector() -> None:
    assert is_per_interval_or_per_day_vector("PER_DAY_Vector")
    assert is_per_interval_or_per_day_vector("PER_INTVL_Vector")
    assert not is_per_interval_or_per_day_vector("per_day_Vector")
    assert not is_per_interval_or_per_day_vector("per_intvl_Vector")
    assert not is_per_interval_or_per_day_vector("vector")


def test_get_cumulative_vector_name() -> None:
    assert get_cumulative_vector_name("PER_DAY_FOPT") == "FOPT"
    assert get_cumulative_vector_name("PER_INTVL_FOPT") == "FOPT"

    assert get_cumulative_vector_name("PER_DAY_FOPR") == "FOPR"
    assert get_cumulative_vector_name("PER_INTVL_FOPR") == "FOPR"

    # Expect ValueError when verifying vector not starting with "PER_DAY_" or "PER_INTVL_"
    try:
        get_cumulative_vector_name("Test_vector")
        pytest.fail('Expected retrieving of cumulative vector name for "Test_vector"')
    except ValueError as err:
        assert (
            f"{err}"
            == 'Expected "Test_vector" to be a vector calculated from cumulative!'
        )


def test_create_per_day_vector_name() -> None:
    assert create_per_day_vector_name("Vector") == "PER_DAY_Vector"
    assert create_per_day_vector_name("FOPT") == "PER_DAY_FOPT"
    assert create_per_day_vector_name("FOPS") == "PER_DAY_FOPS"


def test_create_per_interval_vector_name() -> None:
    assert create_per_interval_vector_name("Vector") == "PER_INTVL_Vector"
    assert create_per_interval_vector_name("FOPT") == "PER_INTVL_FOPT"
    assert create_per_interval_vector_name("FOPS") == "PER_INTVL_FOPS"


def test_datetime_to_intervalstr() -> None:
    # Verify early return (ignore mypy)
    assert datetime_to_intervalstr(None, Frequency.WEEKLY) is None  # type: ignore

    test_date = datetime.datetime(2021, 11, 12, 13, 37)
    assert datetime_to_intervalstr(test_date, Frequency.DAILY) == "2021-11-12"
    assert datetime_to_intervalstr(test_date, Frequency.WEEKLY) == "2021-W45"
    assert datetime_to_intervalstr(test_date, Frequency.MONTHLY) == "2021-11"
    assert datetime_to_intervalstr(test_date, Frequency.QUARTERLY) == "2021-Q4"
    assert datetime_to_intervalstr(test_date, Frequency.YEARLY) == "2021"

    # Verify invalid frequency - i.e. isoformat!
    assert datetime_to_intervalstr(test_date, None) == "2021-11-12T13:37:00"  # type: ignore



import datetime

import pandas as pd
import pytest
from pandas._testing import assert_frame_equal
from webviz_subsurface_components.VectorCalculatorWrapper import (
    ExpressionInfo,
    VariableVectorMapInfo,
)

from webviz_subsurface._utils.dataframe_utils import make_date_column_datetime_object

# pylint: disable = line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.derived_vectors_accessor.derived_ensemble_vectors_accessor_impl import (
    DerivedEnsembleVectorsAccessorImpl,
)

from ..mocks.derived_vectors_accessor_ensemble_summary_provider_mock import (
    EnsembleSummaryProviderMock,
)

# *******************************************************************
#####################################################################
#
# CONFIGURE TESTDATA
#
#####################################################################
# *******************************************************************

# fmt: off
INPUT_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "A", "B"],
    data = [
        [datetime.datetime(2000,1,1),  1, 1.0,   50.0  ],
        [datetime.datetime(2000,2,1),  1, 2.0,   100.0 ],
        [datetime.datetime(2000,3,1),  1, 3.0,   150.0 ],
        [datetime.datetime(2000,1,1),  2, 6.0,   300.0 ],
        [datetime.datetime(2000,2,1),  2, 7.0,   400.0 ],
        [datetime.datetime(2000,3,1),  2, 8.0,   500.0 ],
        [datetime.datetime(2000,1,1),  4, 11.0,  1000.0],
        [datetime.datetime(2000,2,1),  4, 12.0,  1200.0],
        [datetime.datetime(2000,3,1),  4, 13.0,  1400.0],
    ]
)

# PER_INTVL_ calc for col "B" input df
EXPECTED_PER_INTVL_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "PER_INTVL_B"],
    data = [
        [datetime.datetime(2000,1,1),  1, 50.0 ],
        [datetime.datetime(2000,2,1),  1, 50.0 ],
        [datetime.datetime(2000,3,1),  1, 0.0  ],
        [datetime.datetime(2000,1,1),  2, 100.0],
        [datetime.datetime(2000,2,1),  2, 100.0],
        [datetime.datetime(2000,3,1),  2, 0.0  ],
        [datetime.datetime(2000,1,1),  4, 200.0],
        [datetime.datetime(2000,2,1),  4, 200.0],
        [datetime.datetime(2000,3,1),  4, 0.0  ],
    ]
)

# Sum of col "A" and "B" in input df
EXPECTED_SUM_A_AND_B_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "Sum A and B"],
    data = [
        [datetime.datetime(2000,1,1),  1, 51.0  ],
        [datetime.datetime(2000,2,1),  1, 102.0 ],
        [datetime.datetime(2000,3,1),  1, 153.0 ],
        [datetime.datetime(2000,1,1),  2, 306.0 ],
        [datetime.datetime(2000,2,1),  2, 407.0 ],
        [datetime.datetime(2000,3,1),  2, 508.0 ],
        [datetime.datetime(2000,1,1),  4, 1011.0],
        [datetime.datetime(2000,2,1),  4, 1212.0],
        [datetime.datetime(2000,3,1),  4, 1413.0],
    ]
)
make_date_column_datetime_object(INPUT_DF)
make_date_column_datetime_object(EXPECTED_PER_INTVL_DF)
make_date_column_datetime_object(EXPECTED_SUM_A_AND_B_DF)

# Dates AFTER year 2262!
AFTER_2262_DATES = pd.Series(
    [
        datetime.datetime(2265,1,1),
        datetime.datetime(2265,2,1),
        datetime.datetime(2265,3,1),
        datetime.datetime(2265,1,1),
        datetime.datetime(2265,2,1),
        datetime.datetime(2265,3,1),
        datetime.datetime(2265,1,1),
        datetime.datetime(2265,2,1),
        datetime.datetime(2265,3,1),
    ]
)
# NOTE: datetime.datetime after year 2262 is not converted to pd.Timestamp, thus
# no need to make date column datetime object
INPUT_AFTER_2262_DF = INPUT_DF.copy()
INPUT_AFTER_2262_DF["DATE"] = AFTER_2262_DATES
EXPECTED_INVTL_AFTER_2262_DF = EXPECTED_PER_INTVL_DF.copy()
EXPECTED_INVTL_AFTER_2262_DF["DATE"] = AFTER_2262_DATES
EXPECTED_SUM_A_AND_B_AFTER_2262_DF = EXPECTED_SUM_A_AND_B_DF.copy()
EXPECTED_SUM_A_AND_B_AFTER_2262_DF["DATE"] = AFTER_2262_DATES

# fmt: on
TEST_EXPRESSION = ExpressionInfo(
    name="Sum A and B",
    expression="x+y",
    id="TestId",
    variableVectorMap=[
        VariableVectorMapInfo(variableName="x", vectorName=["A"]),
        VariableVectorMapInfo(variableName="y", vectorName=["B"]),
    ],
    isValid=True,
    isDeletable=False,
)

TEST_ACCESSOR = DerivedEnsembleVectorsAccessorImpl(
    name="Test accessor",
    provider=EnsembleSummaryProviderMock(INPUT_DF),
    vectors=["A", "B", "PER_INTVL_B", "Sum A and B"],
    expressions=[TEST_EXPRESSION],
    resampling_frequency=None,
)

TEST_AFTER_2262_ACCESSOR = DerivedEnsembleVectorsAccessorImpl(
    name="Test after 2262 accessor",
    provider=EnsembleSummaryProviderMock(INPUT_AFTER_2262_DF),
    vectors=["A", "B", "PER_INTVL_B", "Sum A and B"],
    expressions=[TEST_EXPRESSION],
    resampling_frequency=None,
)

TEST_EMPTY_ACCESSOR = DerivedEnsembleVectorsAccessorImpl(
    name="Empty provider accessor",
    provider=EnsembleSummaryProviderMock(pd.DataFrame()),
    vectors=["A", "B", "PER_INTVL_B", "Sum A and B"],
    expressions=None,
    resampling_frequency=None,
)


# *******************************************************************
#####################################################################
#
# UNIT TESTS
#
#####################################################################
# *******************************************************************

TEST_STATUS_CASES = [
    pytest.param(TEST_ACCESSOR, True),
    pytest.param(TEST_AFTER_2262_ACCESSOR, True),
    pytest.param(TEST_EMPTY_ACCESSOR, False),
]
TEST_GET_VECTOR_CASES = [
    pytest.param(TEST_ACCESSOR, INPUT_DF),
    pytest.param(TEST_AFTER_2262_ACCESSOR, INPUT_AFTER_2262_DF),
]
TEST_CREATE_PER_INTVL_PER_DAY_VECTOR_CASES = [
    pytest.param(TEST_ACCESSOR, EXPECTED_PER_INTVL_DF),
    pytest.param(TEST_AFTER_2262_ACCESSOR, EXPECTED_INVTL_AFTER_2262_DF),
]
TEST_CREATE_CALCULATED_VECTOR_CASES = [
    pytest.param(TEST_ACCESSOR, EXPECTED_SUM_A_AND_B_DF),
    pytest.param(TEST_AFTER_2262_ACCESSOR, EXPECTED_SUM_A_AND_B_AFTER_2262_DF),
]


@pytest.mark.parametrize("test_accessor, expected_state", TEST_STATUS_CASES)
def test_has_provider_vectors(
    test_accessor: DerivedEnsembleVectorsAccessorImpl, expected_state: bool
) -> None:
    assert test_accessor.has_provider_vectors() == expected_state


@pytest.mark.parametrize("test_accessor, expected_state", TEST_STATUS_CASES)
def test_has_per_interval_and_per_day_vectors(
    test_accessor: DerivedEnsembleVectorsAccessorImpl, expected_state: bool
) -> None:
    assert test_accessor.has_per_interval_and_per_day_vectors() == expected_state


@pytest.mark.parametrize("test_accessor, expected_state", TEST_STATUS_CASES)
def test_has_vector_calculator_expressions(
    test_accessor: DerivedEnsembleVectorsAccessorImpl, expected_state: bool
) -> None:
    assert test_accessor.has_vector_calculator_expressions() == expected_state


@pytest.mark.parametrize("test_accessor, expected_df", TEST_GET_VECTOR_CASES)
def test_get_provider_vectors(
    test_accessor: DerivedEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    assert_frame_equal(expected_df, test_accessor.get_provider_vectors_df())


@pytest.mark.parametrize("test_accessor, expected_df", TEST_GET_VECTOR_CASES)
def test_get_provider_vectors_filter_realizations(
    test_accessor: DerivedEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    # Filter realizations
    expected_reals_df = (
        expected_df.loc[expected_df["REAL"].isin([1, 4])]
        .reset_index()
        .drop("index", axis=1)
    )

    test_df = test_accessor.get_provider_vectors_df(realizations=[1, 4])

    assert_frame_equal(expected_reals_df, test_df)
    assert list(set(test_df["REAL"].values)) == [1, 4]


@pytest.mark.parametrize(
    "test_accessor, expected_df", TEST_CREATE_PER_INTVL_PER_DAY_VECTOR_CASES
)
def test_create_per_interval_and_per_day_vectors_df(
    test_accessor: DerivedEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    assert_frame_equal(
        expected_df, test_accessor.create_per_interval_and_per_day_vectors_df()
    )


@pytest.mark.parametrize(
    "test_accessor, expected_df", TEST_CREATE_PER_INTVL_PER_DAY_VECTOR_CASES
)
def test_create_per_interval_and_per_day_vectors_df_filter_realizations(
    test_accessor: DerivedEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    # Filter realizations
    expected_reals_df = (
        expected_df.loc[expected_df["REAL"].isin([1, 2])]
        .reset_index()
        .drop("index", axis=1)
    )

    test_df = test_accessor.create_per_interval_and_per_day_vectors_df(
        realizations=[1, 2]
    )

    assert_frame_equal(expected_reals_df, test_df)
    assert list(set(test_df["REAL"].values)) == [1, 2]


@pytest.mark.parametrize(
    "test_accessor, expected_df", TEST_CREATE_CALCULATED_VECTOR_CASES
)
def test_create_calculated_vectors_df(
    test_accessor: DerivedEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    assert_frame_equal(expected_df, test_accessor.create_calculated_vectors_df())


@pytest.mark.parametrize(
    "test_accessor, expected_df", TEST_CREATE_CALCULATED_VECTOR_CASES
)
def test_create_calculated_vectors_df_filter_realizations(
    test_accessor: DerivedEnsembleVectorsAccessorImpl, expected_df: pd.DataFrame
) -> None:
    # Filter realizations
    expected_reals_df = (
        expected_df.loc[expected_df["REAL"].isin([2, 4])]
        .reset_index()
        .drop("index", axis=1)
    )

    test_df = test_accessor.create_calculated_vectors_df(realizations=[2, 4])

    assert_frame_equal(expected_reals_df, test_df)
    assert list(set(test_df["REAL"].values)) == [2, 4]



########################################
#
#  Copyright (C) 2020-     Equinor ASA
#
########################################

import sys
import warnings
from typing import Any, Dict, List

import numpy as np
import pandas as pd

# opm and ecl2df are only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    import ecl2df
    from opm.io.ecl import EclFile
except ImportError:
    pass

from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from .eclipse_init_io.pvt_gas import Gas
from .eclipse_init_io.pvt_oil import Oil
from .eclipse_init_io.pvt_water import Water
from .fmu_input import load_csv, load_ensemble_set


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def filter_pvt_data_frame(
    data_frame: pd.DataFrame, drop_ensemble_duplicates: bool = False
) -> pd.DataFrame:
    # pylint: disable=too-many-branches
    data_frame = data_frame.rename(str.upper, axis="columns").rename(
        columns={
            "TYPE": "KEYWORD",
            "RS": "GOR",
            "RSO": "GOR",
            "R": "GOR",
            "RV": "OGR",
        }
    )
    data_frame = data_frame.fillna(0)
    if "GOR" in data_frame.columns and "OGR" in data_frame.columns:
        data_frame["RATIO"] = data_frame["GOR"] + data_frame["OGR"]
    elif "GOR" in data_frame.columns:
        data_frame["RATIO"] = data_frame["GOR"]
    elif "OGR" in data_frame.columns:
        data_frame["RATIO"] = data_frame["OGR"]

    columns = [
        "ENSEMBLE",
        "REAL",
        "PVTNUM",
        "KEYWORD",
        "RATIO",
        "PRESSURE",
        "PRESSURE_UNIT",
        "VOLUMEFACTOR",
        "VOLUMEFACTOR_UNIT",
        "VISCOSITY",
        "VISCOSITY_UNIT",
        "DENSITY",
        "DENSITY_UNIT",
        "RATIO_UNIT",
    ]

    if not "RATIO" in data_frame.columns:
        raise ValueError(
            "The dataframe must contain a column for the ratio (OGR, GOR, R, RV, RS)."
        )
    if not "VOLUMEFACTOR_UNIT" in data_frame.columns:
        data_frame["VOLUMEFACTOR_UNIT"] = "rm³/sm³"
    if not "PRESSURE_UNIT" in data_frame.columns:
        data_frame["PRESSURE_UNIT"] = "bar"
    if not "VISCOSITY_UNIT" in data_frame.columns:
        data_frame["VISCOSITY_UNIT"] = "cP"
    if not "DENSITY_UNIT" in data_frame.columns:
        data_frame["DENSITY_UNIT"] = "kg/m³"
    if not "RATIO_UNIT" in data_frame.columns:
        data_frame["RATIO_UNIT"] = "Scm³/Scm³"

    if not "DENSITY" in data_frame.columns:
        data_frame = calculate_densities(data_frame)

    data_frame = data_frame[columns]

    stored_data_frames: List[pd.DataFrame] = []
    cleaned_data_frame = data_frame.iloc[0:0]
    columns_subset = data_frame.columns.difference(["REAL", "ENSEMBLE"])

    for ens, ens_data_frame in data_frame.groupby("ENSEMBLE"):
        ens_merged_dataframe = data_frame.iloc[0:0]
        for _rno, realization_data_frame in ens_data_frame.groupby("REAL"):
            if ens_merged_dataframe.empty:
                ens_merged_dataframe = pd.concat(
                    [ens_merged_dataframe, realization_data_frame], ignore_index=True
                )
            else:
                ens_merged_dataframe = (
                    pd.concat([ens_merged_dataframe, realization_data_frame])
                    .drop_duplicates(
                        subset=columns_subset,
                        keep="first",
                    )
                    .reset_index(drop=True)
                )

        if ens_merged_dataframe["REAL"].nunique() > 1:
            warnings.warn(
                (
                    f"There are variations in PVT between realizations in ensemble {ens}. "
                    "This is currently not supported. Only keeping data for realization "
                    f"{ens_merged_dataframe['REAL'].iloc[0]}."
                ),
                UserWarning,
            )
            ens_merged_dataframe = ens_merged_dataframe[
                ens_merged_dataframe["REAL"] == ens_merged_dataframe["REAL"].iloc[0]
            ]
        if drop_ensemble_duplicates:
            data_frame_stored = False
            for stored_data_frame in stored_data_frames:
                if all(
                    stored_data_frame[columns_subset]
                    == ens_merged_dataframe[columns_subset]
                ):
                    data_frame_stored = True
                    break
            if data_frame_stored:
                continue
            stored_data_frames.append(ens_merged_dataframe)
        cleaned_data_frame = cleaned_data_frame.append(ens_merged_dataframe)

    return cleaned_data_frame


def calculate_densities(data_frame: pd.DataFrame) -> pd.DataFrame:
    oil_density = data_frame.loc[
        data_frame["KEYWORD"] == "DENSITY", "OILDENSITY"
    ].values[0]
    gas_density = data_frame.loc[
        data_frame["KEYWORD"] == "DENSITY", "GASDENSITY"
    ].values[0]
    water_density = data_frame.loc[
        data_frame["KEYWORD"] == "DENSITY", "WATERDENSITY"
    ].values[0]

    def calculate_density(keyword: str, ratio: float, volume_factor: float) -> float:
        density = 0.0
        if keyword == "PVTO":
            density = (oil_density + ratio * gas_density) / volume_factor
        elif keyword == "PVDO":
            density = oil_density / volume_factor
        elif keyword == "PVTG":
            density = (gas_density + ratio * oil_density) / volume_factor
        elif keyword == "PVDG":
            density = gas_density / volume_factor
        elif keyword == "PVCDO":
            density = oil_density / volume_factor
        elif keyword == "PVTW":
            density = water_density / volume_factor
        return density

    data_frame["DENSITY"] = data_frame.apply(
        lambda row: calculate_density(
            row["KEYWORD"], row["RATIO"], row["VOLUMEFACTOR"]
        ),
        axis=1,
    )
    return data_frame


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def load_pvt_csv(
    ensemble_paths: dict,
    csv_file: str,
    ensemble_set_name: str = "EnsembleSet",
    drop_ensemble_duplicates: bool = False,
) -> pd.DataFrame:
    return filter_pvt_data_frame(
        load_csv(ensemble_paths, csv_file, ensemble_set_name), drop_ensemble_duplicates
    )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def load_pvt_dataframe(
    ensemble_paths: Dict[str, str],
    ensemble_set_name: str = "EnsembleSet",
    use_init_file: bool = False,
    drop_ensemble_duplicates: bool = False,
) -> pd.DataFrame:
    # pylint: disable=too-many-statements

    def check_if_ecl2df_is_installed() -> None:
        # If ecl2df is not loaded, this machine is probably not
        # running Linux and the modules are not available.
        if "ecl2df" not in sys.modules:
            raise ModuleNotFoundError(
                "Your operating system does not support opening and reading"
                " Eclipse files. An empty data frame will be returned and your"
                " plot will therefore not show any data points. Please specify"
                " a relative path to a PVT CSV file in your PvtPlot settings"
                " to display PVT data anyways."
            )

    def ecl2df_pvt_data_frame(kwargs: Any) -> pd.DataFrame:
        check_if_ecl2df_is_installed()
        return ecl2df.pvt.df(kwargs["realization"].get_eclfiles())

    def init_to_pvt_data_frame(kwargs: Any) -> pd.DataFrame:
        # pylint: disable-msg=too-many-locals
        # pylint: disable=too-many-branches
        check_if_ecl2df_is_installed()
        ecl_init_file = EclFile(
            kwargs["realization"].get_eclfiles().get_initfile().get_filename()
        )

        # Keep the original unit system
        oil = Oil.from_ecl_init_file(ecl_init_file, True)
        gas = Gas.from_ecl_init_file(ecl_init_file, True)
        water = Water.from_ecl_init_file(ecl_init_file, True)

        column_pvtnum: List[int] = []
        column_ratio: List[float] = []
        column_volume_factor: List[float] = []
        column_volume_factor_unit: List[str] = []
        column_pressure: List[float] = []
        column_pressure_unit: List[str] = []
        column_viscosity: List[float] = []
        column_viscosity_unit: List[str] = []
        column_density: List[float] = []
        column_density_unit: List[str] = []
        column_ratio_unit: List[str] = []
        column_keyword: List[str] = []

        ratios: np.ndarray = np.zeros(21)
        pressures: np.ndarray = np.zeros(21)

        if oil and not oil.is_dead_oil_const_compr():
            (pressure_min, pressure_max) = oil.range_independent(0)
        elif gas:
            (pressure_min, pressure_max) = gas.range_independent(0)
        else:
            raise NotImplementedError("Missing PVT data")

        for pressure_step in range(0, 20 + 1):
            pressures[pressure_step] = pressure_min + pressure_step / 20.0 * (
                pressure_max - pressure_min
            )
            ratios[pressure_step] = 0.0

        if oil:
            if oil.is_live_oil():
                keyword = "PVTO"
            elif oil.is_dead_oil():
                keyword = "PVDO"
            elif oil.is_dead_oil_const_compr():
                keyword = "PVCDO"
            else:
                raise NotImplementedError(
                    "The PVT property type of oil is not implemented."
                )

            for region_index, region in enumerate(oil.regions()):
                if oil.is_dead_oil_const_compr():
                    column_pvtnum.extend([region_index + 1 for _ in pressures])
                    column_keyword.extend([keyword for _ in pressures])
                    column_ratio.extend(ratios)
                    column_pressure.extend(pressures)
                    column_pressure_unit.extend(
                        [oil.pressure_unit() for _ in pressures]
                    )
                    column_volume_factor.extend(
                        region.formation_volume_factor(ratios, pressures)
                    )
                    column_volume_factor_unit.extend(
                        [oil.formation_volume_factor_unit() for _ in pressures]
                    )
                    column_viscosity.extend(region.viscosity(ratios, pressures))
                    column_viscosity_unit.extend(
                        [oil.viscosity_unit() for _ in pressures]
                    )
                    column_density.extend(region.density(ratios, pressures))
                    column_density_unit.extend([oil.density_unit() for _ in pressures])
                    column_ratio_unit.extend([oil.ratio_unit() for _ in pressures])

                else:
                    (ratio, pressure) = (
                        region.get_keys(),
                        region.get_independents(),
                    )
                    column_pvtnum.extend([region_index + 1 for _ in pressure])
                    column_keyword.extend([keyword for _ in pressure])
                    column_ratio.extend(ratio)
                    column_pressure.extend(pressure)
                    column_pressure_unit.extend([oil.pressure_unit() for _ in pressure])
                    column_volume_factor.extend(
                        region.formation_volume_factor(ratio, pressure)
                    )
                    column_volume_factor_unit.extend(
                        [oil.formation_volume_factor_unit() for _ in pressure]
                    )
                    column_viscosity.extend(region.viscosity(ratio, pressure))
                    column_viscosity_unit.extend(
                        [oil.viscosity_unit() for _ in pressure]
                    )
                    column_density.extend(region.density(ratio, pressure))
                    column_density_unit.extend([oil.density_unit() for _ in pressure])
                    column_ratio_unit.extend([oil.ratio_unit() for _ in pressure])

        if gas:
            if gas.is_wet_gas():
                keyword = "PVTG"
            elif gas.is_dry_gas():
                keyword = "PVDG"
            else:
                raise NotImplementedError(
                    "The PVT property type of gas is not implemented."
                )

            for region_index, region in enumerate(gas.regions()):
                if gas.is_wet_gas():
                    (pressure, ratio) = (
                        region.get_keys(),
                        region.get_independents(),
                    )
                else:
                    (ratio, pressure) = (
                        region.get_keys(),
                        region.get_independents(),
                    )
                column_pvtnum.extend([region_index + 1 for _ in pressure])
                column_keyword.extend([keyword for _ in pressure])
                column_ratio.extend(ratio)
                column_pressure.extend(pressure)
                column_pressure_unit.extend([gas.pressure_unit() for _ in pressure])
                column_volume_factor.extend(
                    region.formation_volume_factor(ratio, pressure)
                )
                column_volume_factor_unit.extend(
                    [gas.formation_volume_factor_unit() for _ in pressure]
                )
                column_viscosity.extend(region.viscosity(ratio, pressure))
                column_viscosity_unit.extend([gas.viscosity_unit() for _ in pressure])

                column_density.extend(region.density(ratio, pressure))
                column_density_unit.extend([gas.density_unit() for _ in pressure])
                column_ratio_unit.extend([gas.ratio_unit() for _ in pressure])

        if water:
            for region_index, region in enumerate(water.regions()):
                column_pvtnum.extend([region_index + 1 for _ in pressures])
                column_keyword.extend(["PVTW" for _ in pressures])
                column_ratio.extend(ratios)
                column_pressure.extend(pressures)
                column_pressure_unit.extend([water.pressure_unit() for _ in pressures])
                column_volume_factor.extend(
                    region.formation_volume_factor(ratios, pressures)
                )
                column_volume_factor_unit.extend(
                    [water.formation_volume_factor_unit() for _ in pressures]
                )
                column_viscosity.extend(region.viscosity(ratios, pressures))
                column_viscosity_unit.extend(
                    [water.viscosity_unit() for _ in pressures]
                )
                column_density.extend(region.density(ratio, pressures))
                column_density_unit.extend([water.density_unit() for _ in pressures])
                column_ratio_unit.extend(["" for _ in pressures])

        data_frame = pd.DataFrame(
            {
                "PVTNUM": column_pvtnum,
                "KEYWORD": column_keyword,
                "R": column_ratio,
                "PRESSURE": column_pressure,
                "PRESSURE_UNIT": column_pressure_unit,
                "VOLUMEFACTOR": column_volume_factor,
                "VOLUMEFACTOR_UNIT": column_volume_factor_unit,
                "VISCOSITY": column_viscosity,
                "VISCOSITY_UNIT": column_viscosity_unit,
                "DENSITY": column_density,
                "DENSITY_UNIT": column_density_unit,
                "RATIO_UNIT": column_ratio_unit,
            }
        )

        return data_frame

    return filter_pvt_data_frame(
        load_ensemble_set(ensemble_paths, ensemble_set_name).apply(
            init_to_pvt_data_frame if use_init_file else ecl2df_pvt_data_frame
        ),
        drop_ensemble_duplicates,
    )



from .ensemble_timeseries_datamodel import ProviderTimeSeriesDataModel
from .property_statistics_model import PropertyStatisticsModel
from .simulation_timeseries_model import SimulationTimeSeriesModel



# pylint: disable=no-name-in-module
from webviz_config.plugins import SegyViewer


def test_segy_viewer(dash_duo, app, shared_settings, testdata_folder) -> None:
    plugin = SegyViewer(
        app,
        shared_settings["HM_SETTINGS"],
        segyfiles=[
            testdata_folder
            / "01_drogon_ahm"
            / "realization-0"
            / "iter-0"
            / "share"
            / "results"
            / "seismic"
            / "seismic--amplitude_depth--20180701_20180101.segy"
        ],
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class PlotType(StrEnum):
    CROSSPLOT = "crossplot"
    ERROR_BOXPLOT = "error-boxplot"


class PlotTypeSettings(SettingsGroupABC):
    class Ids(StrEnum):
        PLOT_TYPE = "plot-type"

    def __init__(self) -> None:
        super().__init__("Plot Type")

    def layout(self) -> List[Component]:
        return [
            wcc.RadioItems(
                id=self.register_component_unique_id(self.Ids.PLOT_TYPE),
                options=[
                    {
                        "label": "CrossPlot",
                        "value": PlotType.CROSSPLOT,
                    },
                    {
                        "label": "Error BoxPlot",
                        "value": PlotType.ERROR_BOXPLOT,
                    },
                ],
                value=PlotType.CROSSPLOT,
            ),
        ]



import io
import json

import numpy as np
import pandas as pd
import pytest

from webviz_subsurface._models.well_attributes_model import WellAttributesModel

LOAD_DATA_FUNCTION = (
    "webviz_subsurface._models.well_attributes_model.WellAttributesModel._load_data"
)


def test_simplest_possible_input(mocker):
    mocker.patch(
        LOAD_DATA_FUNCTION,
        lambda self: io.BytesIO(json.dumps({"version": "0.1", "wells": []}).encode()),
    )
    well_attr = WellAttributesModel("ens_name", "ens_path", "file_name")
    assert well_attr.data == {}


def test_not_implemented_version(mocker):
    mocker.patch(
        LOAD_DATA_FUNCTION,
        lambda self: io.BytesIO(json.dumps({"version": "0.2", "wells": []}).encode()),
    )
    with pytest.raises(NotImplementedError):
        WellAttributesModel("ens_name", "ens_path", "file_name")


def test_object_properties(mocker):
    mock_data = {
        "version": "0.1",
        "wells": [
            {
                "alias": {"eclipse": "OP_1"},
                "attributes": {
                    "attr1": "value1",
                    "attr2": "value2",
                },
                "name": "OP_1",
            },
            {
                "alias": {"eclipse": "OP_2"},
                "attributes": {
                    "attr1": "value1",
                },
                "name": "OP_2",
            },
        ],
    }
    mocker.patch(
        LOAD_DATA_FUNCTION,
        lambda self: io.BytesIO(json.dumps(mock_data).encode()),
    )
    well_attr = WellAttributesModel("ens_name", "ens_path", "file_name")

    # Test categories
    assert set(well_attr.categories) == {"attr1", "attr2"}

    # Test data
    assert well_attr.data == {
        "OP_1": {
            "attr1": "value1",
            "attr2": "value2",
        },
        "OP_2": {
            "attr1": "value1",
        },
    }

    # Test category_dict (This will replace nan with Undefined)
    assert well_attr.category_dict == {
        "attr1": ["value1"],
        "attr2": ["value2", "Undefined"],
    }

    # Test dataframe
    pd.testing.assert_frame_equal(
        well_attr.dataframe,
        pd.DataFrame(
            columns=["WELL", "attr1", "attr2"],
            data=[["OP_1", "value1", "value2"], ["OP_2", "value1", np.nan]],
        ),
    )

    # Test melted dataframe
    df_melted = well_attr.dataframe_melted
    assert df_melted.shape[0] == 4
    assert df_melted.shape[1] == 3
    assert set(df_melted["WELL"]) == {"OP_1", "OP_2"}
    assert set(df_melted["CATEGORY"]) == {"attr1", "attr2"}
    assert set(df_melted["VALUE"]) == {"value1", "value2", np.nan}

    # This is the expected dataframe, but I'm not able to test it with
    # assert_frame_equal because the row order of the melted dataframe is
    # randomly changing.
    #
    # exp_df_melted = pd.DataFrame(
    #     columns=["WELL", "CATEGORY", "VALUE"],
    #     data=[
    #         ["OP_1", "attr1", "value1"],
    #         ["OP_2", "attr1", "value1"],
    #         ["OP_1", "attr2", "value2"],
    #         ["OP_2", "attr2", np.nan],
    #     ],
    # )
    # pd.testing.assert_frame_equal(df, exp_df, check_like=True)



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

TESTS_REQUIRE = [
    "bandit",
    "black>=22.1",
    "dash[testing]",
    "flaky",
    "isort",
    "mypy",
    "pylint<=2.13.9",  # Locked due to https://github.com/equinor/webviz-subsurface/issues/1052
    "pytest-mock",
    "pytest-xdist",
    "selenium>=3.141",
    "types-pkg-resources",
    "types-pyyaml",
]

# pylint: disable=line-too-long
setup(
    name="webviz-subsurface",
    description="Webviz config plugins for subsurface data",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/equinor/webviz-subsurface",
    author="R&T Equinor",
    packages=find_packages(exclude=["tests"]),
    package_data={
        "webviz_subsurface": [
            "_abbreviations/abbreviation_data/*.json",
            "_assets/css/*.css",
            "_assets/js/*.js",
            "ert_jobs/config_jobs/*",
        ]
    },
    entry_points={
        "webviz_config_plugins": [
            "AssistedHistoryMatchingAnalysis = webviz_subsurface.plugins._assisted_history_matching_analysis:AssistedHistoryMatchingAnalysis",
            "BhpQc = webviz_subsurface.plugins._bhp_qc:BhpQc",
            "CO2Leakage = webviz_subsurface.plugins._co2_leakage:CO2Leakage",
            "DiskUsage = webviz_subsurface.plugins._disk_usage:DiskUsage",
            "EXPERIMENTALGridViewerFMU = webviz_subsurface.plugins._grid_viewer_fmu:EXPERIMENTALGridViewerFMU",
            "GroupTree = webviz_subsurface.plugins._group_tree:GroupTree",
            "HistoryMatch = webviz_subsurface.plugins._history_match:HistoryMatch",
            "HorizonUncertaintyViewer = webviz_subsurface.plugins._horizon_uncertainty_viewer:HorizonUncertaintyViewer",
            "InplaceVolumes = webviz_subsurface.plugins._inplace_volumes:InplaceVolumes",
            "InplaceVolumesOneByOne = webviz_subsurface.plugins._inplace_volumes_onebyone:InplaceVolumesOneByOne",
            "LinePlotterFMU = webviz_subsurface.plugins._line_plotter_fmu.line_plotter_fmu:LinePlotterFMU",
            "MapViewerFMU = webviz_subsurface.plugins._map_viewer_fmu:MapViewerFMU",
            "MorrisPlot = webviz_subsurface.plugins._morris_plot:MorrisPlot",
            "ParameterAnalysis = webviz_subsurface.plugins._parameter_analysis:ParameterAnalysis",
            "ParameterCorrelation = webviz_subsurface.plugins._parameter_correlation:ParameterCorrelation",
            "ParameterDistribution = webviz_subsurface.plugins._parameter_distribution:ParameterDistribution",
            "ParameterParallelCoordinates = webviz_subsurface.plugins._parameter_parallel_coordinates:ParameterParallelCoordinates",
            "ParameterResponseCorrelation = webviz_subsurface.plugins._parameter_response_correlation:ParameterResponseCorrelation",
            "ProdMisfit = webviz_subsurface.plugins._prod_misfit:ProdMisfit",
            "PropertyStatistics = webviz_subsurface.plugins._property_statistics:PropertyStatistics",
            "PvtPlot = webviz_subsurface.plugins._pvt_plot:PvtPlot",
            "RelativePermeability = webviz_subsurface.plugins._relative_permeability:RelativePermeability",
            "ReservoirSimulationTimeSeries = webviz_subsurface.plugins._reservoir_simulation_timeseries:ReservoirSimulationTimeSeries",
            "ReservoirSimulationTimeSeriesOneByOne = webviz_subsurface.plugins._reservoir_simulation_timeseries_onebyone:ReservoirSimulationTimeSeriesOneByOne",
            "ReservoirSimulationTimeSeriesRegional = webviz_subsurface.plugins._reservoir_simulation_timeseries_regional:ReservoirSimulationTimeSeriesRegional",
            "RftPlotter = webviz_subsurface.plugins._rft_plotter:RftPlotter",
            "RunningTimeAnalysisFMU = webviz_subsurface.plugins._running_time_analysis_fmu:RunningTimeAnalysisFMU",
            "SegyViewer = webviz_subsurface.plugins._segy_viewer:SegyViewer",
            "SeismicMisfit = webviz_subsurface.plugins._seismic_misfit:SeismicMisfit",
            "SimulationTimeSeries = webviz_subsurface.plugins._simulation_time_series:SimulationTimeSeries",
            "StructuralUncertainty = webviz_subsurface.plugins._structural_uncertainty:StructuralUncertainty",
            "SubsurfaceMap = webviz_subsurface.plugins._subsurface_map:SubsurfaceMap",
            "SurfaceViewerFMU = webviz_subsurface.plugins._surface_viewer_fmu:SurfaceViewerFMU",
            "SurfaceWithGridCrossSection = webviz_subsurface.plugins._surface_with_grid_cross_section:SurfaceWithGridCrossSection",
            "SurfaceWithSeismicCrossSection = webviz_subsurface.plugins._surface_with_seismic_cross_section :SurfaceWithSeismicCrossSection",
            "SwatinitQC = webviz_subsurface.plugins._swatinit_qc:SwatinitQC",
            "TornadoPlotterFMU = webviz_subsurface.plugins._tornado_plotter_fmu:TornadoPlotterFMU",
            "VfpAnalysis = webviz_subsurface.plugins._vfp_analysis:VfpAnalysis",
            "VolumetricAnalysis = webviz_subsurface.plugins._volumetric_analysis:VolumetricAnalysis",
            "WellAnalysis = webviz_subsurface.plugins._well_analysis:WellAnalysis",
            "WellCompletion = webviz_subsurface.plugins._well_completion:WellCompletion",
            "WellCompletions = webviz_subsurface.plugins._well_completions:WellCompletions",
            "WellCrossSection = webviz_subsurface.plugins._well_cross_section:WellCrossSection",
            "WellCrossSectionFMU = webviz_subsurface.plugins._well_cross_section_fmu:WellCrossSectionFMU",
            "WellLogViewer = webviz_subsurface.plugins._well_log_viewer:WellLogViewer",
        ],
        "console_scripts": ["smry2arrow_batch=webviz_subsurface.smry2arrow_batch:main"],
    },
    install_requires=[
        "flask-caching<1.11",  # https://github.com/equinor/webviz-subsurface/issues/1043
        "dash>=2.0.0",
        "dash_bootstrap_components>=0.10.3",
        "dash-daq>=0.5.0",
        "defusedxml>=0.6.0",
        "ecl2df>=0.15.0; sys_platform=='linux'",
        "fmu-ensemble>=1.2.3",
        "fmu-tools>=1.8",
        "geojson>=2.5.0",
        "jsonschema>=3.2.0",
        "opm>=2020.10.1; sys_platform=='linux'",
        "pandas>=1.1.5",
        "pillow>=6.1",
        "pyarrow>=5.0.0",
        "pyjwt>=2.6.0",
        "pyscal>=0.7.5",
        "scipy>=1.2",
        "statsmodels>=0.12.1",  # indirect dependency through https://plotly.com/python/linear-fits/
        "xtgeo>=2.20.0",
        "vtk>=9.2.2",
        "webviz-config",
        "webviz-core-components>=0.6",
        "webviz-subsurface-components>=0.4.14",
    ],
    extras_require={"tests": TESTS_REQUIRE},
    setup_requires=["setuptools_scm~=3.2"],
    python_requires="~=3.8",
    use_scm_version=True,
    zip_safe=False,
    project_urls={
        "Documentation": "https://equinor.github.io/webviz-subsurface",
        "Download": "https://pypi.org/project/webviz-subsurface/",
        "Source": "https://github.com/equinor/webviz-subsurface",
        "Tracker": "https://github.com/equinor/webviz-subsurface/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Environment :: Web Environment",
        "Framework :: Dash",
        "Framework :: Flask",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)
